#ifndef MTZ_HPP
#define MTZ_HPP

// MTZ file reading
// NOTE: input stream MUST be opened in binary mode
//
// All functions are inline to allow header-only implementation.

#include <string>
#include <fstream>
#include <cmath>
#include <array>
#include <limits>
#include <cstring>
#include <cstdint>

static_assert(std::numeric_limits<unsigned char>::digits == 8 &&
                  sizeof(float) == 4,
              "float type must be 32 bits");

namespace crystallography {

#if defined(__BIG_ENDIAN__)
constexpr short MtzHostFloatFormat = 1;
#else
constexpr short MtzHostFloatFormat = 4;
#endif

constexpr auto MtzReflectionStartPosition = 80;

static float readfloat(std::istream& stream, bool swap_bytes);
static int readint(std::istream& stream, bool swap_bytes);
static std::streampos header_start_position(std::istream& mtzstream);
static bool byteswap_needed(std::istream& mtzstream);
static bool seek_to_header(std::istream& stream, const char* key);

// checks whether the file appears to be an MTZ file
inline bool mtz_is_valid(std::istream& stream) {
  if (stream.fail())
    throw std::ios_base::failure("unable to open the MTZ file");
  stream.exceptions(std::ifstream::badbit | std::ios_base::failbit);
  stream.seekg(0);
  char buf[4];
  stream.read(buf, 4);
  return (strncmp(buf, "MTZ ", 4) == 0);
}

// number of reflections
inline int mtz_nrefl(std::istream& stream) {
  auto number = 0;
  if (!seek_to_header(stream, "NCOL")) {
    return 0;
  }
  std::string skip;
  stream >> skip >> number;
  return number;
}

// space group number
inline int mtz_sgnum(std::istream& stream) {
  auto number = 0;
  if (!seek_to_header(stream, "SYMINF")) {
    return 0;
  }
  std::string skip;
  stream >> skip >> skip >> skip >> number;
  return number;
}

// number of symmetry operations
inline int mtz_nsymm(std::istream& stream) {
  auto number = 0;
  if (!seek_to_header(stream, "SYMINF")) {
    return 0;
  }
  stream >> number;
  return number;
}

inline bool is_column_header(std::string& label) {
  return (label.rfind("COLU", 0) == 0 || label == "COL");
}

// finds a data column with a label starting with the given string or returns an
// empty string
inline std::string
mtz_first_column_label_starting_with(std::istream& stream,
                                     const char* labelstart) {
  auto nextkey = std::string{};
  auto str = std::string{};
  auto pos = header_start_position(stream);
  while (nextkey != "END") {
    stream.seekg(pos);
    stream >> nextkey;
    if (is_column_header(nextkey)) {
      stream >> str;
      if (str.rfind(labelstart, 0) == 0)
        return str;
    }
    pos += 80;
  }
  return std::string{};
}

inline std::string mtz_first_column_of_type(std::istream& stream, int type) {
  auto nextkey = std::string{};
  auto str = std::string{};
  auto pos = header_start_position(stream);
  while (nextkey != "END") {
    stream.seekg(pos);
    stream >> nextkey;
    if (is_column_header(nextkey)) {
      stream >> str;
      // ISSUE: a label with internal spaces
      // would cause this to fail.
      auto typestr = std::string{};
      stream >> typestr;
      if (typestr[0] == type)
        return str;
    }
    pos += 80;
  }
  return std::string{};
}

// returns the position of data column with the given label or -1
inline int mtz_column_position(std::istream& stream, const std::string& label) {
  auto nextkey = std::string{};
  auto str = std::string{};
  auto pos = header_start_position(stream);
  auto i = 0;
  while (nextkey != "END") {
    stream.seekg(pos);
    stream >> nextkey;
    if (is_column_header(nextkey)) {
      stream >> str;
      if (str == label)
        return i;
      else
        ++i;
    }
    pos += 80;
  }
  return -1;
}

inline int mtz_column_type(std::istream& stream, const std::string& label) {
  auto nextkey = std::string{};
  auto str = std::string{};
  auto pos = header_start_position(stream);
  auto i = 0;
  while (nextkey != "END") {
    stream.seekg(pos);
    stream >> nextkey;
    if (is_column_header(nextkey)) {
      stream >> str;
      if (str == label) {
        auto type = std::string{};
        // ISSUE: a label with internal spaces
        // would cause this to fail.
        stream >> type;
        return type[0];
      } else
        ++i;
    }
    pos += 80;
  }
  return -1;
}

// number of data columns for each reflection
inline int mtz_ncol(std::istream& stream) {
  auto ncol = 0;
  if (!seek_to_header(stream, "NCOL")) {
    return 0;
  }
  stream >> ncol;
  return ncol;
}

// retrieve the data items in the specified columns
inline void mtz_data_columns(std::istream& stream, int record_number,
                             int* column_positions, float* data_items,
                             int nitems) {
  auto swap_bytes = byteswap_needed(stream);
  auto ncolumns = mtz_ncol(stream);
  auto record_position =
      MtzReflectionStartPosition + record_number * sizeof(float) * ncolumns;
  for (auto i = 0; i < nitems; ++i) {
    stream.seekg(record_position +
                 (std::streampos)(column_positions[i] * sizeof(float)));
    data_items[i] = readfloat(stream, swap_bytes);
  }
}

// retrieve unit cell constants
inline void mtz_cell_constants(std::istream& stream,
                               std::array<double, 6>& param) {
  // ISSUE: CELL is deprecated, so DCELL should be preferred, and it's
  // associated dataset ID should be matched with that of the data
  // column that was selected, but that will require a change in the
  // order of loading so that the dataset ID is known.
  if (seek_to_header(stream, "CELL") || seek_to_header(stream, "DCELL")) {
    stream >> param[0] >> param[1] >> param[2] >> param[3] >> param[4] >>
        param[5];
  }
}

inline bool byteswap_needed(std::istream& mtzstream) {
  char buf[4];
  mtzstream.seekg(2 * sizeof(float));
  mtzstream.read(buf, 4);
  return (buf[0] >> 4) == MtzHostFloatFormat ? false : true;
}

inline std::streampos header_start_position(std::istream& mtzstream) {
  auto swap_bytes = byteswap_needed(mtzstream);
  mtzstream.seekg(sizeof(float));
  // the position of the text of the first header, after the keyword,
  // is what is encoded, so adjust back
  return sizeof(float) * readint(mtzstream, swap_bytes) - 4;
}

inline bool seek_to_header(std::istream& stream, const char* key) {
  auto nextkey = std::string{};
  auto str = std::string{};
  auto pos = header_start_position(stream);
  while (nextkey != "END") {
    stream.seekg(pos);
    stream >> nextkey;
    if (nextkey == key) {
      return true;
    }
    pos += 80;
  }
  return false;
}

inline std::int32_t readint(std::istream& stream, bool swap_bytes) {
  union {
    char c[4];
    std::int32_t i;
  } buf;
  stream.read(buf.c, 4);
  if (swap_bytes) {
    std::swap(buf.c[0], buf.c[3]);
    std::swap(buf.c[1], buf.c[2]);
  }
  return buf.i;
}

inline float readfloat(std::istream& stream, bool swap_bytes) {
  union {
    char c[4];
    float f;
  } buf;
  stream.read(buf.c, 4);
  if (swap_bytes) {
    std::swap(buf.c[0], buf.c[3]);
    std::swap(buf.c[1], buf.c[2]);
  }
  return buf.f;
}
}

#endif
