/*
 Copyright (c) 2015 Charles R. Kissinger
 Released under the MIT license: http://opensource.org/licenses/MIT
 */

#ifndef RFACTOR_HPP
#define RFACTOR_HPP

#include <cmath>
#include <algorithm>

template <class T, class ResultT = double>
class rfactor {
public:
  rfactor(){};
  void add(T f1, T f2, T weight = 1);
  ResultT operator()();

private:
  ResultT sumfo = 0;
  ResultT sumdiff = 0;
};

template <class T, class ResultT>
inline void rfactor<T, ResultT>::add(T f1, T f2, T weight) {
  sumfo += weight * f1;
  sumdiff += weight * std::abs(f1 - f2);
}

template <class T, class ResultT>
inline ResultT rfactor<T, ResultT>::operator()() {
  return sumdiff / sumfo;
}

template <class Inp1, class Inp2, class ResultT>
ResultT calc_rfactor(Inp1 first, Inp1 last, Inp2 first2) {
  rfactor<typename std::iterator_traits<Inp1>::value_type, ResultT> r;
  while (first != last)
    r.add(*first++, *first2++);
  return r();
}

template <class Inp1, class Inp2, class Inp3, class ResultT>
ResultT calc_rfactor(Inp1 first, Inp1 last, Inp2 first2, Inp3 firstw) {
  rfactor<typename std::iterator_traits<Inp1>::value_type, ResultT> r;
  while (first != last)
    r.add(*first++, *first2++, *firstw++);
  return r();
}

template <class Inp1, class Inp2, class Inp3, class ResultT = double>
ResultT calc_scaled_rfactor(Inp1 first, Inp1 last, Inp2 first2, Inp3 firstw) {
  auto sumfo = std::accumulate(first, last, ResultT{0});
  auto sumfc = std::accumulate(first2, first2 + (last - first), ResultT{0});
  auto scale = sumfo / sumfc;
  rfactor<typename std::iterator_traits<Inp1>::value_type, ResultT> r;
  while (first != last)
    r.add(*first++, *first2++ * scale, *firstw++);
  return r();
}

template <class T, class ResultT = double>
class scaled_rfactor {
public:
  scaled_rfactor(){};
  void add(T f1, T f2, T weight = 1);
  ResultT operator()();

private:
  std::vector<ResultT> fo;
  std::vector<ResultT> fc;
  std::vector<ResultT> weights;
};

template <class T, class ResultT>
inline void scaled_rfactor<T, ResultT>::add(T f1, T f2, T weight) {
  fo.push_back(f1);
  fc.push_back(f2);
  weights.push_back(weight);
}

template <class T, class ResultT>
inline ResultT scaled_rfactor<T, ResultT>::operator()() {
  return calc_scaled_rfactor(std::begin(fo), std::end(fo), std::begin(fc),
                             std::begin(weights));
}

#endif
