/*
 Copyright (c) 2015 Charles R. Kissinger
 */

#ifndef SPACEGROUP_HPP
#define SPACEGROUP_HPP

#include <vector>
#include <string>
#include <memory>
#include <stdexcept>
#include <array>
#include <istream>
#include <cstring>

namespace crystallography {

// This class stores information associated with a crystallographic
// space group. Instances of the class are retrieved using the static
// from_number() member function and are immutable.
class spacegroup {

public:
  // The following are left unimplemented to preserve SpaceGroup singletons
  spacegroup& operator=(const spacegroup&) = delete;
  spacegroup(const spacegroup&) = delete;
  spacegroup(const spacegroup&&) = delete;

  // Retrieves a SpaceGroup instance corresponding to the specified
  // space group number
  // Throws a std::invalid_argument exception if the int argument is less
  // than 1 or greater than 230
  static spacegroup& from_number(int n);

  static int symbol_to_number(std::string symbol) noexcept;

  // Returns the space group number
  int number() const noexcept { return num; }

  // Returns the compact symbolic name for the space group
  const std::string& symbol() const noexcept { return name; }

  // Returns a vector containing all Hermann-Maugin symbols
  // for this space group
  const std::vector<std::string>& hm_symbols() const noexcept {
    return hm_names;
  }

  // Returns the symbol for the point group to which the space group
  // belongs
  const std::string& point_group() const noexcept { return pointgroup; }

  // Returns the name of the crystal class to which
  // the space group belongs (e.g., orthorhombic)
  const std::string& crystal_system() const noexcept { return cryst_system; }

  // Returns a character indicating the lattice type (e.g., P,C,I)
  char lattice_type() const noexcept { return hm_names[0][0]; }

  // Returns the set of symmetry operators for this space group
  const std::vector<std::array<std::array<double, 4>, 4>>&
  symmetry_operators() const noexcept {
    return operators;
  }

  // Returns the number of symmetry operations for this space group
  int num_symm_ops() const noexcept { return (int)operators.size(); }

  // Returns the number of primitive symmetry operators
  // These are placed first in the set of matrices returned by symm_op().
  int num_primitive_symm_ops() const noexcept {
    return num_primitive_operators;
  }

  // Returns a 4 x 4 transformation matrix representing the symmetry operation
  const std::array<std::array<double, 4>, 4>& symm_op(int n) const noexcept {
    return operators[n];
  }

  // Returns the number of alternative origins
  int origin_count() const noexcept { return (int)origins.size(); }

  // Returns a point representing the origin
  std::array<double, 3> get_origin(int n) const noexcept { return origins[n]; }

  // Returns true if the a axis is polar (origin undefined)
  bool is_a_axis_polar() const noexcept { return polar_axes[0]; }

  // Returns true if the b axis is polar (origin undefined)
  bool is_b_axis_polar() const noexcept { return polar_axes[1]; }

  // Returns true if the c axis is polar (origin undefined)
  bool is_c_axis_polar() const noexcept { return polar_axes[2]; }

  // Returns true if the space group is polar
  bool is_polar() const noexcept {
    return is_a_axis_polar() || is_b_axis_polar() || is_c_axis_polar();
  }

  // number of space groups
  static constexpr auto NumberOfSpaceGroups = 230;

private:
  // private constructor taking a space group number
  spacegroup(int num);

  static const char* const* textreps();

  // space group number
  int num;

  // space group symbol
  std::string name;

  // Hermann-Maugin symbols
  std::vector<std::string> hm_names;

  // name of crystal system
  std::string cryst_system;

  // point group symbol
  std::string pointgroup;

  // number of primitive symmetry operators
  int num_primitive_operators;

  // transformation matrices for each symmetry operation
  std::vector<std::array<std::array<double, 4>, 4>> operators;

  // the set of alternate origins
  std::vector<std::array<double, 3>> origins;

  // flags indicating whether each axis is polar
  // true if polar, false if not
  std::array<bool, 3> polar_axes;
};

// return array of character strings containing information to
// create a spacegroup instance for each known space group
// textreps[0] and textreps[231] are unused
// The static array of strings is wrapped in a function call
// for better behavior in a header file.
inline const char* const* spacegroup::textreps() {
  // string representations of space groups
  // each string contains symbol, HM symbol, point group, crystal system,
  // number of primitive symm ops, number of symm ops,
  // 3x4 transformation matrices with translations in twelfths,
  // and alternate origins

  // Using constexpr here causes the program to crash when compiled with
  // Visual Studio 2015, Update 1
  static const char* reps[] = {
      "", "P1 (P 1) 1 triclinic 1 1 1 0 0 0 1 0 0 0 1 0 0 0 "
          " -999.9 -999.9 -999.9",
      "P-1 (P -1) 1bar triclinic 2 2 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P2 (P 1 2 1) 2 monoclinic 2 2 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      " 0.0 -999.9 0.0 0.0 -999.9 0.5 0.5 -999.9 0.0 0.5 -999.9 0.5",
      "P21 (P 1 21 1) 2 monoclinic 2 2 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 0 "
      " 0.0 -999.9 0.0 0.0 -999.9 0.5 0.5 -999.9 0.0 0.5 -999.9 0.5",
      "C2 (C 1 2 1) 2 monoclinic 2 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      " 0.0 -999.9 0.0 0.0 -999.9 0.5 0.5 -999.9 0.0 0.5 -999.9 0.5",
      "Pm (P 1 m 1) m monoclinic 2 2 1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9",
      "Pc (P 1 c 1) m monoclinic 2 2 1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9",
      "Cm (C 1 m 1) m monoclinic 2 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9",
      "Cc (C 1 c 1) m monoclinic 2 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9",
      "P2/m (P 1 2/m 1) 2/m monoclinic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P21/m (P 1 21/m 1) 2/m monoclinic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "C2/m (C 1 2/m 1) 2/m monoclinic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P2/c (P 1 2/c 1) 2/m monoclinic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P21/c (P 1 21/c 1) 2/m monoclinic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "C2/c (C 1 2/c 1) 2/m monoclinic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P222 (P 2 2 2) 222 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P2221 (P 2 2 21) 222 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P21212 (P 21 21 2) 222 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P212121 (P 21 21 21) 222 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "C2221 (C 2 2 21) 222 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "C222 (C 2 2 2) 222 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "F222 (F 2 2 2) 222 orthorhombic 4 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I222 (I 2 2 2) 222 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I212121 (I 21 21 21) 222 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pmm2 (P m m 2) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pmc21 (P m c 21) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pcc2 (P c c 2) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pma2 (P m a 2) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 1 6 0 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pca21 (P c a 21) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pnc2 (P n c 2) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pmn21 (P m n 21) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pba2 (P b a 2) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pna21 (P n a 21) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pnn2 (P n n 2) mm2 orthorhombic 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Cmm2 (C m m 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Cmc21 (C m c 21) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Ccc2 (C c c 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Amm2 (A m m 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Abm2 (A b m 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 1 0 6 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Ama2 (A m a 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 1 6 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Aba2 (A b a 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Fmm2 (F m m 2) mm2 orthorhombic 4 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Fdd2 (F d d 2) mm2 orthorhombic 4 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 3 3 3 "
      "-1 0 0 0 1 0 0 0 1 3 3 3 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 3 9 9 "
      "-1 0 0 0 1 0 0 0 1 3 9 9 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 9 3 9 "
      "-1 0 0 0 1 0 0 0 1 9 3 9 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 9 9 3 "
      "-1 0 0 0 1 0 0 0 1 9 9 3 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Imm2 (I m m 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Iba2 (I b a 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Ima2 (I m a 2) mm2 orthorhombic 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 1 6 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      " 0.0 0.0 -999.9 0.0 0.5 -999.9 0.5 0.0 -999.9 0.5 0.5 -999.9",
      "Pmmm (P 2/m 2/m 2/m,P m m m) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pnnn (P 2/n 2/n 2/n,P n n n) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pccm (P 2/c 2/c 2/m,P c c m) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pban (P 2/b 2/a 2/n,P b a n) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pmma (P 21/m 2/m 2/a,P m m a) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 6 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 6 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pnna (P 2/n 21/n 2/a,P n n a) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pmna (P 2/m 2/n 21/a,P m n a) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pcca (P 21/c 2/c 2/a,P c c a) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pbam (P 21/b 21/a 2/m,P b a m) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pccn (P 21/c 21/c 2/n,P c c n) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pbcm (P 2/b 21/c 21/m,P b c m) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pnnm (P 21/n 21/n 2/m,P n n m) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pmmn (P 21/m 21/m 2/n,P m m n) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pbcn (P 21/b 2/c 21/n,P b c n) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pbca (P 21/b 21/c 21/a,P b c a) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 "
      "0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Pnma (P 21/n 21/m 21/a,P n m a) mmm orthorhombic 8 8 1 0 0 0 1 0 0 0 1 "
      "0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Cmcm (C 2/m 2/c 21/m,C m c m) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Cmca (C 2/m 2/c 21/a,C m c a) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Cmmm (C 2/m 2/m 2/m,C m m m) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Cccm (C 2/c 2/c 2/m,C c c m) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Cmma (C 2/m 2/m 2/a,C m m a) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 6 0 "
      "1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Ccca (C 2/c 2/c 2/a,C c c a) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Fmmm (F 2/m 2/m 2/m,F m m m) mmm orthorhombic 8 32 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Fddd (F 2/d 2/d 2/d,F d d d) mmm orthorhombic 8 32 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 3 3 3 "
      "1 0 0 0 1 0 0 0 -1 3 3 3 "
      "1 0 0 0 -1 0 0 0 1 3 3 3 "
      "-1 0 0 0 1 0 0 0 1 3 3 3 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 3 9 9 "
      "1 0 0 0 1 0 0 0 -1 3 9 9 "
      "1 0 0 0 -1 0 0 0 1 3 9 9 "
      "-1 0 0 0 1 0 0 0 1 3 9 9 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 9 3 9 "
      "1 0 0 0 1 0 0 0 -1 9 3 9 "
      "1 0 0 0 -1 0 0 0 1 9 3 9 "
      "-1 0 0 0 1 0 0 0 1 9 3 9 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 9 9 3 "
      "1 0 0 0 1 0 0 0 -1 9 9 3 "
      "1 0 0 0 -1 0 0 0 1 9 9 3 "
      "-1 0 0 0 1 0 0 0 1 9 9 3 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Immm (I 2/m 2/m 2/m,I m m m) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Ibam (I 2/b 2/a 2/m,I b a m) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Ibca (I 21/b 21/c 21/a,I b c a) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 "
      "0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 0 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "Imma (I 21/m 21/m 21/a,I m m a) mmm orthorhombic 8 16 1 0 0 0 1 0 0 0 1 "
      "0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 6 0 "
      "1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.0 0.5 0.0 0.0 0.5 0.5 0.5 0.0 0.0 0.5 0.0 "
      "0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4 (P 4) 4 tetragonal 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P41 (P 41) 4 tetragonal 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 1 0 0 3 "
      "0 1 0 -1 0 0 0 0 1 0 0 9 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P42 (P 42) 4 tetragonal 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P43 (P 43) 4 tetragonal 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 1 0 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 0 3 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "I4 (I 4) 4 tetragonal 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "I41 (I 41) 4 tetragonal 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 3 "
      "0 1 0 -1 0 0 0 0 1 6 0 9 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 6 3 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P-4 (P -4) 4bar tetragonal 4 4 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I-4 (I -4) 4bar tetragonal 4 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/m (P 4/m) 4/m tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/m (P 42/m) 4/m tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/n (P 4/n) 4/m tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 0 "
      "0 1 0 -1 0 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/n (P 42/n) 4/m tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I4/m (I 4/m) 4/m tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I41/a (I 41/a) 4/m tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 3 "
      "0 1 0 -1 0 0 0 0 1 6 0 9 "
      "-1 0 0 0 -1 0 0 0 -1 0 6 3 "
      "1 0 0 0 1 0 0 0 -1 6 0 9 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 6 3 "
      "-1 0 0 0 -1 0 0 0 -1 6 0 9 "
      "1 0 0 0 1 0 0 0 -1 0 6 3 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P422 (P 4 2 2) 422 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4212 (P 4 21 2) 422 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 0 "
      "0 1 0 -1 0 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4122 (P 41 2 2) 422 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 1 0 0 3 "
      "0 1 0 -1 0 0 0 0 1 0 0 9 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 9 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 3 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P41212 (P 41 21 2) 422 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 3 "
      "0 1 0 -1 0 0 0 0 1 6 6 9 "
      "-1 0 0 0 1 0 0 0 -1 6 6 3 "
      "1 0 0 0 -1 0 0 0 -1 6 6 9 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4222 (P 42 2 2) 422 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42212 (P 42 21 2) 422 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4322 (P 43 2 2) 422 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 1 0 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 0 3 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 3 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 9 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P43212 (P 43 21 2) 422 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 9 "
      "0 1 0 -1 0 0 0 0 1 6 6 3 "
      "-1 0 0 0 1 0 0 0 -1 6 6 9 "
      "1 0 0 0 -1 0 0 0 -1 6 6 3 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I422 (I 4 2 2) 422 tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I4122 (I 41 2 2) 422 tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 3 "
      "0 1 0 -1 0 0 0 0 1 6 0 9 "
      "-1 0 0 0 1 0 0 0 -1 6 0 9 "
      "1 0 0 0 -1 0 0 0 -1 0 6 3 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 6 3 "
      "-1 0 0 0 1 0 0 0 -1 0 6 3 "
      "1 0 0 0 -1 0 0 0 -1 6 0 9 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4mm (P 4 m m) 4mm tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P4bm (P 4 b m) 4mm tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P42cm (P 42 c m) 4mm tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P42nm (P 42 n m) 4mm tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.0",
      "P4cc (P 4 c c) 4mm tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P4nc (P 4 n c) 4mm tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P42mc (P 42 m c) 4mm tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P42bc (P 42 b c) 4mm tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "I4mm (I 4 m m) 4mm tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.0",
      "I4cm (I 4 c m) 4mm tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "I41md (I 41 m d) 4mm tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 3 "
      "0 1 0 -1 0 0 0 0 1 6 0 9 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 0 6 3 "
      "0 1 0 1 0 0 0 0 1 6 0 9 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 6 3 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 6 0 9 "
      "0 1 0 1 0 0 0 0 1 0 6 3 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "I41cd (I 41 c d) 4mm tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 3 "
      "0 1 0 -1 0 0 0 0 1 6 0 9 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 0 6 9 "
      "0 1 0 1 0 0 0 0 1 6 0 3 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 6 3 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 6 0 3 "
      "0 1 0 1 0 0 0 0 1 0 6 9 "
      " 0.0 0.0 -999.9 0.5 0.5 -999.9",
      "P-42m (P -4 2 m) 4bar2m tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P-42c (P -4 2 c) 4bar2m tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P-421m (P -4 21 m) 4bar2m tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P-421c (P -4 21 c) 4bar2m tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P-4m2 (P -4 m 2) 4barm2 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P-4c2 (P -4 c 2) 4barm2 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P-4b2 (P -4 b 2) 4barm2 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P-4n2 (P -4 n 2) 4barm2 tetragonal 8 8 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I-4m2 (I -4 m 2) 4barm2 tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I-4c2 (I -4 c 2) 4barm2 tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I-42m (I -4 2 m) 4bar2m tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I-42d (I -4 2 d) 4bar2m tetragonal 8 16 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 9 "
      "1 0 0 0 -1 0 0 0 -1 6 0 9 "
      "0 -1 0 -1 0 0 0 0 1 6 0 9 "
      "0 1 0 1 0 0 0 0 1 6 0 9 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 3 "
      "1 0 0 0 -1 0 0 0 -1 0 6 3 "
      "0 -1 0 -1 0 0 0 0 1 0 6 3 "
      "0 1 0 1 0 0 0 0 1 0 6 3 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/mmm (P 4/m 2/m 2/m,P4/m m m) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/mcc (P 4/m 2/c 2/c,P4/m c c) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/nbm (P 4/n 2/b 2/m,P4/n b m) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "0 1 0 -1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 1 0 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/nnc (P 4/n 2/n 2/c,P4/n n c) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/mbm (P 4/m 21/b 2/m,P4/m b m) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/mnc (P 4/m 21/n 2/c,P4/m n c) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/nmm (P 4/n 21/m 2/m,P4/n m m) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 0 "
      "0 1 0 -1 0 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P4/ncc (P 4/n 2/c 2/c,P4/n c c) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 0 "
      "0 1 0 -1 0 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/mmc (P 42/m 2/m 2/c,P42/m m c) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 "
      "0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/mcm (P 42/m 2/c 2/m,P42/m c m) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 "
      "0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/nbc (P 42/n 2/b 2/c,P42/n b c) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 "
      "0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/nnm (P 42/n 2/n 2/m,P42/n n m) 4/mmm tetragonal 16 16 1 0 0 0 1 0 0 "
      "0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/mbc (P 42/m 21/b 2/c,P42/m b c) 4/mmm tetragonal 16 16 1 0 0 0 1 0 "
      "0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/mnm (P 42/m 21/n 2/m,P42/m n m) 4/mmm tetragonal 16 16 1 0 0 0 1 0 "
      "0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/nmc (P 42/n 21/m 2/c,P42/n m c) 4/mmm tetragonal 16 16 1 0 0 0 1 0 "
      "0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P42/ncm (P 42/n 21/c 2/m,P42/n c m) 4/mmm tetragonal 16 16 1 0 0 0 1 0 "
      "0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I4/mmm (I 4/m 2/m 2/m,I4/m m m) 4/mmm tetragonal 16 32 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I4/mcm (I 4/m 2/c 2/m,I4/m c m) 4/mmm tetragonal 16 32 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I41/amd (I 41/a 2/m 2/d,I41/a m d) 4/mmm tetragonal 16 32 1 0 0 0 1 0 0 "
      "0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 3 "
      "0 1 0 -1 0 0 0 0 1 6 0 9 "
      "-1 0 0 0 1 0 0 0 -1 6 0 9 "
      "1 0 0 0 -1 0 0 0 -1 0 6 3 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 6 3 "
      "1 0 0 0 1 0 0 0 -1 6 0 9 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 6 0 9 "
      "0 1 0 1 0 0 0 0 1 0 6 3 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 6 3 "
      "-1 0 0 0 1 0 0 0 -1 0 6 3 "
      "1 0 0 0 -1 0 0 0 -1 6 0 9 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 0 9 "
      "1 0 0 0 1 0 0 0 -1 0 6 3 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 0 6 3 "
      "0 1 0 1 0 0 0 0 1 6 0 9 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "I41/acd (I 41/a 2/c 2/d,I41/a c d) 4/mmm tetragonal 16 32 1 0 0 0 1 0 0 "
      "0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 3 "
      "0 1 0 -1 0 0 0 0 1 6 0 9 "
      "-1 0 0 0 1 0 0 0 -1 6 0 3 "
      "1 0 0 0 -1 0 0 0 -1 0 6 9 "
      "0 1 0 1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 6 3 "
      "1 0 0 0 1 0 0 0 -1 6 0 9 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 6 0 3 "
      "0 1 0 1 0 0 0 0 1 0 6 9 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 0 9 "
      "0 1 0 -1 0 0 0 0 1 0 6 3 "
      "-1 0 0 0 1 0 0 0 -1 0 6 9 "
      "1 0 0 0 -1 0 0 0 -1 6 0 3 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 0 9 "
      "1 0 0 0 1 0 0 0 -1 0 6 3 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 0 6 9 "
      "0 1 0 1 0 0 0 0 1 6 0 3 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.5 0.5 0.0 0.5 0.5 0.5",
      "P3 (P 3) 3 trigonal 3 3 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.3333333333333333 0.6666666666666667 -999.9 "
      "0.6666666666666667 0.3333333333333333 -999.9",
      "P31 (P 31) 3 trigonal 3 3 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 4 "
      "-1 1 0 -1 0 0 0 0 1 0 0 8 "
      " 0.0 0.0 -999.9 0.3333333333333333 0.6666666666666667 -999.9 "
      "0.6666666666666667 0.3333333333333333 -999.9",
      "P32 (P 32) 3 trigonal 3 3 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 8 "
      "-1 1 0 -1 0 0 0 0 1 0 0 4 "
      " 0.0 0.0 -999.9 0.3333333333333333 0.6666666666666667 -999.9 "
      "0.6666666666666667 0.3333333333333333 -999.9",
      "R3 (R 3) 3 trigonal 3 9 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 8 4 4 "
      "0 -1 0 1 -1 0 0 0 1 8 4 4 "
      "-1 1 0 -1 0 0 0 0 1 8 4 4 "
      "1 0 0 0 1 0 0 0 1 4 8 8 "
      "0 -1 0 1 -1 0 0 0 1 4 8 8 "
      "-1 1 0 -1 0 0 0 0 1 4 8 8 "
      " 0.0 0.0 -999.9 0.3333333333333333 0.6666666666666667 -999.9 "
      "0.6666666666666667 0.3333333333333333 -999.9",
      "P-3 (P -3) 3bar trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "R-3 (R -3) 3bar trigonal 6 18 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 8 4 4 "
      "0 -1 0 1 -1 0 0 0 1 8 4 4 "
      "-1 1 0 -1 0 0 0 0 1 8 4 4 "
      "-1 0 0 0 -1 0 0 0 -1 8 4 4 "
      "0 1 0 -1 1 0 0 0 -1 8 4 4 "
      "1 -1 0 1 0 0 0 0 -1 8 4 4 "
      "1 0 0 0 1 0 0 0 1 4 8 8 "
      "0 -1 0 1 -1 0 0 0 1 4 8 8 "
      "-1 1 0 -1 0 0 0 0 1 4 8 8 "
      "-1 0 0 0 -1 0 0 0 -1 4 8 8 "
      "0 1 0 -1 1 0 0 0 -1 4 8 8 "
      "1 -1 0 1 0 0 0 0 -1 4 8 8 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P312 (P 3 1 2) 312 trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.3333333333333333 0.6666666666666667 0.0 "
      "0.3333333333333333 0.6666666666666667 0.5 0.6666666666666667 "
      "0.3333333333333333 0.0 0.6666666666666667 0.3333333333333333 0.5",
      "P321 (P 3 2 1) 321 trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P3112 (P 31 1 2) 312 trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 4 "
      "-1 1 0 -1 0 0 0 0 1 0 0 8 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 8 "
      "-1 1 0 0 1 0 0 0 -1 0 0 4 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.3333333333333333 0.6666666666666667 0.0 "
      "0.3333333333333333 0.6666666666666667 0.5 0.6666666666666667 "
      "0.3333333333333333 0.0 0.6666666666666667 0.3333333333333333 0.5",
      "P3121 (P 31 2 1) 321 trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 4 "
      "-1 1 0 -1 0 0 0 0 1 0 0 8 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 8 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 4 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P3212 (P 32 1 2) 312 trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 8 "
      "-1 1 0 -1 0 0 0 0 1 0 0 4 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 4 "
      "-1 1 0 0 1 0 0 0 -1 0 0 8 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.3333333333333333 0.6666666666666667 0.0 "
      "0.3333333333333333 0.6666666666666667 0.5 0.6666666666666667 "
      "0.3333333333333333 0.0 0.6666666666666667 0.3333333333333333 0.5",
      "P3221 (P 32 2 1) 321 trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 8 "
      "-1 1 0 -1 0 0 0 0 1 0 0 4 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 4 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 8 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "R32 (R 3 2) 32 trigonal 6 18 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 8 4 4 "
      "0 -1 0 1 -1 0 0 0 1 8 4 4 "
      "-1 1 0 -1 0 0 0 0 1 8 4 4 "
      "0 1 0 1 0 0 0 0 -1 8 4 4 "
      "1 -1 0 0 -1 0 0 0 -1 8 4 4 "
      "-1 0 0 -1 1 0 0 0 -1 8 4 4 "
      "1 0 0 0 1 0 0 0 1 4 8 8 "
      "0 -1 0 1 -1 0 0 0 1 4 8 8 "
      "-1 1 0 -1 0 0 0 0 1 4 8 8 "
      "0 1 0 1 0 0 0 0 -1 4 8 8 "
      "1 -1 0 0 -1 0 0 0 -1 4 8 8 "
      "-1 0 0 -1 1 0 0 0 -1 4 8 8 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P3m1 (P 3 m 1) 3m1 trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9 0.3333333333333333 0.6666666666666667 -999.9 "
      "0.6666666666666667 0.3333333333333333 -999.9",
      "P31m (P 3 1 m) 31m trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9",
      "P3c1 (P 3 c 1) 3m1 trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9 0.3333333333333333 0.6666666666666667 -999.9 "
      "0.6666666666666667 0.3333333333333333 -999.9",
      "P31c (P 3 1 c) 31m trigonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9",
      "R3m (R 3 m) 3m trigonal 6 18 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 8 4 4 "
      "0 -1 0 1 -1 0 0 0 1 8 4 4 "
      "-1 1 0 -1 0 0 0 0 1 8 4 4 "
      "0 -1 0 -1 0 0 0 0 1 8 4 4 "
      "-1 1 0 0 1 0 0 0 1 8 4 4 "
      "1 0 0 1 -1 0 0 0 1 8 4 4 "
      "1 0 0 0 1 0 0 0 1 4 8 8 "
      "0 -1 0 1 -1 0 0 0 1 4 8 8 "
      "-1 1 0 -1 0 0 0 0 1 4 8 8 "
      "0 -1 0 -1 0 0 0 0 1 4 8 8 "
      "-1 1 0 0 1 0 0 0 1 4 8 8 "
      "1 0 0 1 -1 0 0 0 1 4 8 8 "
      " 0.0 0.0 -999.9 0.3333333333333333 0.6666666666666667 -999.9 "
      "0.6666666666666667 0.3333333333333333 -999.9",
      "R3c (R 3 c) 3m trigonal 6 18 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 8 4 4 "
      "0 -1 0 1 -1 0 0 0 1 8 4 4 "
      "-1 1 0 -1 0 0 0 0 1 8 4 4 "
      "0 -1 0 -1 0 0 0 0 1 8 4 10 "
      "-1 1 0 0 1 0 0 0 1 8 4 10 "
      "1 0 0 1 -1 0 0 0 1 8 4 10 "
      "1 0 0 0 1 0 0 0 1 4 8 8 "
      "0 -1 0 1 -1 0 0 0 1 4 8 8 "
      "-1 1 0 -1 0 0 0 0 1 4 8 8 "
      "0 -1 0 -1 0 0 0 0 1 4 8 2 "
      "-1 1 0 0 1 0 0 0 1 4 8 2 "
      "1 0 0 1 -1 0 0 0 1 4 8 2 "
      " 0.0 0.0 -999.9 0.3333333333333333 0.6666666666666667 -999.9 "
      "0.6666666666666667 0.3333333333333333 -999.9",
      "P-31m (P -3 1 2/m,P -3 1 m) 3bar1m trigonal 12 12 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P-31c (P -3 1 2/c,P -3 1 c) 3bar1m trigonal 12 12 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 1 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 1 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P-3m1 (P -3 2/m 1,P -3 m 1) 3barm1 trigonal 12 12 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P-3c1 (P -3 2/c 1,P -3 c 1) 3barm1 trigonal 12 12 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "R-3m (R -3 2/m,R -3 m) 3barm trigonal 12 36 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 1 8 4 4 "
      "0 -1 0 1 -1 0 0 0 1 8 4 4 "
      "-1 1 0 -1 0 0 0 0 1 8 4 4 "
      "0 1 0 1 0 0 0 0 -1 8 4 4 "
      "1 -1 0 0 -1 0 0 0 -1 8 4 4 "
      "-1 0 0 -1 1 0 0 0 -1 8 4 4 "
      "-1 0 0 0 -1 0 0 0 -1 8 4 4 "
      "0 1 0 -1 1 0 0 0 -1 8 4 4 "
      "1 -1 0 1 0 0 0 0 -1 8 4 4 "
      "0 -1 0 -1 0 0 0 0 1 8 4 4 "
      "-1 1 0 0 1 0 0 0 1 8 4 4 "
      "1 0 0 1 -1 0 0 0 1 8 4 4 "
      "1 0 0 0 1 0 0 0 1 4 8 8 "
      "0 -1 0 1 -1 0 0 0 1 4 8 8 "
      "-1 1 0 -1 0 0 0 0 1 4 8 8 "
      "0 1 0 1 0 0 0 0 -1 4 8 8 "
      "1 -1 0 0 -1 0 0 0 -1 4 8 8 "
      "-1 0 0 -1 1 0 0 0 -1 4 8 8 "
      "-1 0 0 0 -1 0 0 0 -1 4 8 8 "
      "0 1 0 -1 1 0 0 0 -1 4 8 8 "
      "1 -1 0 1 0 0 0 0 -1 4 8 8 "
      "0 -1 0 -1 0 0 0 0 1 4 8 8 "
      "-1 1 0 0 1 0 0 0 1 4 8 8 "
      "1 0 0 1 -1 0 0 0 1 4 8 8 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "R-3c (R -3 2/c,R -3 c) 3barm trigonal 12 36 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      "1 0 0 0 1 0 0 0 1 8 4 4 "
      "0 -1 0 1 -1 0 0 0 1 8 4 4 "
      "-1 1 0 -1 0 0 0 0 1 8 4 4 "
      "0 1 0 1 0 0 0 0 -1 8 4 10 "
      "1 -1 0 0 -1 0 0 0 -1 8 4 10 "
      "-1 0 0 -1 1 0 0 0 -1 8 4 10 "
      "-1 0 0 0 -1 0 0 0 -1 8 4 4 "
      "0 1 0 -1 1 0 0 0 -1 8 4 4 "
      "1 -1 0 1 0 0 0 0 -1 8 4 4 "
      "0 -1 0 -1 0 0 0 0 1 8 4 10 "
      "-1 1 0 0 1 0 0 0 1 8 4 10 "
      "1 0 0 1 -1 0 0 0 1 8 4 10 "
      "1 0 0 0 1 0 0 0 1 4 8 8 "
      "0 -1 0 1 -1 0 0 0 1 4 8 8 "
      "-1 1 0 -1 0 0 0 0 1 4 8 8 "
      "0 1 0 1 0 0 0 0 -1 4 8 2 "
      "1 -1 0 0 -1 0 0 0 -1 4 8 2 "
      "-1 0 0 -1 1 0 0 0 -1 4 8 2 "
      "-1 0 0 0 -1 0 0 0 -1 4 8 8 "
      "0 1 0 -1 1 0 0 0 -1 4 8 8 "
      "1 -1 0 1 0 0 0 0 -1 4 8 8 "
      "0 -1 0 -1 0 0 0 0 1 4 8 2 "
      "-1 1 0 0 1 0 0 0 1 4 8 2 "
      "1 0 0 1 -1 0 0 0 1 4 8 2 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6 (P 6) 6 hexagonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 0 "
      "1 -1 0 1 0 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9",
      "P61 (P 61) 6 hexagonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 4 "
      "-1 1 0 -1 0 0 0 0 1 0 0 8 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 10 "
      "1 -1 0 1 0 0 0 0 1 0 0 2 "
      " 0.0 0.0 -999.9",
      "P65 (P 65) 6 hexagonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 8 "
      "-1 1 0 -1 0 0 0 0 1 0 0 4 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 2 "
      "1 -1 0 1 0 0 0 0 1 0 0 10 "
      " 0.0 0.0 -999.9",
      "P62 (P 62) 6 hexagonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 8 "
      "-1 1 0 -1 0 0 0 0 1 0 0 4 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 8 "
      "1 -1 0 1 0 0 0 0 1 0 0 4 "
      " 0.0 0.0 -999.9",
      "P64 (P 64) 6 hexagonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 4 "
      "-1 1 0 -1 0 0 0 0 1 0 0 8 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 4 "
      "1 -1 0 1 0 0 0 0 1 0 0 8 "
      " 0.0 0.0 -999.9",
      "P63 (P 63) 6 hexagonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 6 "
      "1 -1 0 1 0 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9",
      "P-6 (P -6) 6bar hexagonal 6 6 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.3333333333333333 0.6666666666666667 0.0 "
      "0.3333333333333333 0.6666666666666667 0.5 0.6666666666666667 "
      "0.3333333333333333 0.0 0.6666666666666667 0.3333333333333333 0.5",
      "P6/m (P 6/m) 6/m hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 0 "
      "1 -1 0 1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P63/m (P 63/m) 6/m hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 6 "
      "1 -1 0 1 0 0 0 0 1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 6 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 6 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P622 (P 6 2 2) 622 hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 0 "
      "1 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6122 (P 61 2 2) 622 hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 4 "
      "-1 1 0 -1 0 0 0 0 1 0 0 8 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 10 "
      "1 -1 0 1 0 0 0 0 1 0 0 2 "
      "0 1 0 1 0 0 0 0 -1 0 0 4 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 8 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 10 "
      "-1 1 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 1 -1 0 0 0 -1 0 0 2 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6522 (P 65 2 2) 622 hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 8 "
      "-1 1 0 -1 0 0 0 0 1 0 0 4 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 2 "
      "1 -1 0 1 0 0 0 0 1 0 0 10 "
      "0 1 0 1 0 0 0 0 -1 0 0 8 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 4 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 2 "
      "-1 1 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 1 -1 0 0 0 -1 0 0 10 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6222 (P 62 2 2) 622 hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 8 "
      "-1 1 0 -1 0 0 0 0 1 0 0 4 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 8 "
      "1 -1 0 1 0 0 0 0 1 0 0 4 "
      "0 1 0 1 0 0 0 0 -1 0 0 8 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 4 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 8 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 4 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6422 (P 64 2 2) 622 hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 4 "
      "-1 1 0 -1 0 0 0 0 1 0 0 8 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 4 "
      "1 -1 0 1 0 0 0 0 1 0 0 8 "
      "0 1 0 1 0 0 0 0 -1 0 0 4 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 8 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 4 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 8 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6322 (P 63 2 2) 622 hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 6 "
      "1 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 1 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 1 -1 0 0 0 -1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6mm (P 6 m m) 6mm hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 0 "
      "1 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9",
      "P6cc (P 6 c c) 6mm hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 0 "
      "1 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9",
      "P63cm (P 63 c m) 6mm hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 6 "
      "1 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 -999.9",
      "P63mc (P 63 m c) 6mm hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 6 "
      "1 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 -999.9",
      "P-6m2 (P -6 m 2) 6barm2 hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.3333333333333333 0.6666666666666667 0.0 "
      "0.3333333333333333 0.6666666666666667 0.5 0.6666666666666667 "
      "0.3333333333333333 0.0 0.6666666666666667 0.3333333333333333 0.5",
      "P-6c2 (P -6 c 2) 6barm2 hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 6 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 6 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5 0.3333333333333333 0.6666666666666667 0.0 "
      "0.3333333333333333 0.6666666666666667 0.5 0.6666666666666667 "
      "0.3333333333333333 0.0 0.6666666666666667 0.3333333333333333 0.5",
      "P-62m (P -6 2 m) 6bar2m hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P-62c (P -6 2 c) 6bar2m hexagonal 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 6 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 6 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6/mmm (P 6/m 2/m 2/m,P 6/m m m) 6/mmm hexagonal 24 24 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 0 "
      "1 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P6/mcc (P 6/m 2/c 2/c,P 6/m c c) 6/mmm hexagonal 24 24 1 0 0 0 1 0 0 0 "
      "1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "0 1 0 -1 1 0 0 0 1 0 0 0 "
      "1 -1 0 1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 1 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 1 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P63/mcm (P 63/m 2/c 2/m,P 63/m c m) 6/mmm hexagonal 24 24 1 0 0 0 1 0 0 "
      "0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 6 "
      "1 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 1 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 1 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 6 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "-1 1 0 0 1 0 0 0 1 0 0 6 "
      "1 0 0 1 -1 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 1 0 0 0 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P63/mmc (P 63/m 2/m 2/c,P 63/m m c) 6/mmm hexagonal 24 24 1 0 0 0 1 0 0 "
      "0 1 0 0 0 "
      "0 -1 0 1 -1 0 0 0 1 0 0 0 "
      "-1 1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 6 "
      "0 1 0 -1 1 0 0 0 1 0 0 6 "
      "1 -1 0 1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "1 -1 0 0 -1 0 0 0 -1 0 0 0 "
      "-1 0 0 -1 1 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 1 0 0 1 0 0 0 -1 0 0 6 "
      "1 0 0 1 -1 0 0 0 -1 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 1 0 -1 1 0 0 0 -1 0 0 0 "
      "1 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 6 "
      "-1 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 1 -1 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "-1 1 0 0 1 0 0 0 1 0 0 0 "
      "1 0 0 1 -1 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "1 -1 0 0 -1 0 0 0 1 0 0 6 "
      "-1 0 0 -1 1 0 0 0 1 0 0 6 "
      " 0.0 0.0 0.0 0.0 0.0 0.5",
      "P23 (P 2 3) 23 cubic 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "F23 (F 2 3) 23 cubic 12 48 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "I23 (I 2 3) 23 cubic 12 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 -1 0 0 0 1 0 6 6 6 "
      "0 0 -1 1 0 0 0 -1 0 6 6 6 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 -1 0 0 0 -1 1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "P213 (P 21 3) 23 cubic 12 12 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "I213 (I 21 3) 23 cubic 12 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 0 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 0 0 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Pm-3 (P 2/m -3,P m -3) m3bar cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 1 0 0 0 -1 0 0 0 0 "
      "0 0 1 -1 0 0 0 1 0 0 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 0 "
      "0 1 0 0 0 1 -1 0 0 0 0 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Pn-3 (P 2/n -3,P n -3) m3bar cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 1 0 0 0 -1 0 6 6 6 "
      "0 0 1 -1 0 0 0 1 0 6 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 1 0 0 6 6 6 "
      "0 1 0 0 0 1 -1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Fm-3 (F 2/m -3,F m -3) m3bar cubic 24 96 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 1 0 0 0 -1 0 0 0 0 "
      "0 0 1 -1 0 0 0 1 0 0 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 0 "
      "0 1 0 0 0 1 -1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      "0 0 -1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 1 0 0 0 -1 0 0 6 6 "
      "0 0 1 -1 0 0 0 1 0 0 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 1 0 0 0 6 6 "
      "0 1 0 0 0 1 -1 0 0 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 1 0 0 0 -1 0 6 0 6 "
      "0 0 1 -1 0 0 0 1 0 6 0 6 "
      "0 -1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 1 0 0 6 0 6 "
      "0 1 0 0 0 1 -1 0 0 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 1 0 0 0 -1 0 6 6 0 "
      "0 0 1 -1 0 0 0 1 0 6 6 0 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 1 0 0 6 6 0 "
      "0 1 0 0 0 1 -1 0 0 6 6 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Fd-3 (F 2/d -3,F d -3) m3bar cubic 24 96 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 3 3 3 "
      "1 0 0 0 1 0 0 0 -1 3 3 3 "
      "1 0 0 0 -1 0 0 0 1 3 3 3 "
      "-1 0 0 0 1 0 0 0 1 3 3 3 "
      "0 0 -1 -1 0 0 0 -1 0 3 3 3 "
      "0 0 -1 1 0 0 0 1 0 3 3 3 "
      "0 0 1 1 0 0 0 -1 0 3 3 3 "
      "0 0 1 -1 0 0 0 1 0 3 3 3 "
      "0 -1 0 0 0 -1 -1 0 0 3 3 3 "
      "0 1 0 0 0 -1 1 0 0 3 3 3 "
      "0 -1 0 0 0 1 1 0 0 3 3 3 "
      "0 1 0 0 0 1 -1 0 0 3 3 3 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 3 9 9 "
      "1 0 0 0 1 0 0 0 -1 3 9 9 "
      "1 0 0 0 -1 0 0 0 1 3 9 9 "
      "-1 0 0 0 1 0 0 0 1 3 9 9 "
      "0 0 -1 -1 0 0 0 -1 0 3 9 9 "
      "0 0 -1 1 0 0 0 1 0 3 9 9 "
      "0 0 1 1 0 0 0 -1 0 3 9 9 "
      "0 0 1 -1 0 0 0 1 0 3 9 9 "
      "0 -1 0 0 0 -1 -1 0 0 3 9 9 "
      "0 1 0 0 0 -1 1 0 0 3 9 9 "
      "0 -1 0 0 0 1 1 0 0 3 9 9 "
      "0 1 0 0 0 1 -1 0 0 3 9 9 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 9 3 9 "
      "1 0 0 0 1 0 0 0 -1 9 3 9 "
      "1 0 0 0 -1 0 0 0 1 9 3 9 "
      "-1 0 0 0 1 0 0 0 1 9 3 9 "
      "0 0 -1 -1 0 0 0 -1 0 9 3 9 "
      "0 0 -1 1 0 0 0 1 0 9 3 9 "
      "0 0 1 1 0 0 0 -1 0 9 3 9 "
      "0 0 1 -1 0 0 0 1 0 9 3 9 "
      "0 -1 0 0 0 -1 -1 0 0 9 3 9 "
      "0 1 0 0 0 -1 1 0 0 9 3 9 "
      "0 -1 0 0 0 1 1 0 0 9 3 9 "
      "0 1 0 0 0 1 -1 0 0 9 3 9 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 9 9 3 "
      "1 0 0 0 1 0 0 0 -1 9 9 3 "
      "1 0 0 0 -1 0 0 0 1 9 9 3 "
      "-1 0 0 0 1 0 0 0 1 9 9 3 "
      "0 0 -1 -1 0 0 0 -1 0 9 9 3 "
      "0 0 -1 1 0 0 0 1 0 9 9 3 "
      "0 0 1 1 0 0 0 -1 0 9 9 3 "
      "0 0 1 -1 0 0 0 1 0 9 9 3 "
      "0 -1 0 0 0 -1 -1 0 0 9 9 3 "
      "0 1 0 0 0 -1 1 0 0 9 9 3 "
      "0 -1 0 0 0 1 1 0 0 9 9 3 "
      "0 1 0 0 0 1 -1 0 0 9 9 3 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Im-3 (I 2/m -3,I m -3) m3bar cubic 24 48 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 1 0 0 0 -1 0 0 0 0 "
      "0 0 1 -1 0 0 0 1 0 0 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 0 "
      "0 1 0 0 0 1 -1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 -1 0 0 0 1 0 6 6 6 "
      "0 0 -1 1 0 0 0 -1 0 6 6 6 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 -1 0 0 0 -1 1 0 0 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 1 0 0 0 -1 0 6 6 6 "
      "0 0 1 -1 0 0 0 1 0 6 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 1 0 0 6 6 6 "
      "0 1 0 0 0 1 -1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Pa-3 (P 21/a -3,P a -3) m3bar cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 1 0 0 0 -1 0 6 0 6 "
      "0 0 1 -1 0 0 0 1 0 0 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 1 0 0 6 6 0 "
      "0 1 0 0 0 1 -1 0 0 6 0 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Ia-3 (I 21/a -3,I a -3) m3bar cubic 24 48 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 1 0 0 0 -1 0 6 0 6 "
      "0 0 1 -1 0 0 0 1 0 0 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 1 0 0 6 6 0 "
      "0 1 0 0 0 1 -1 0 0 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 0 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 0 0 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 0 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 1 0 0 0 1 0 0 0 6 "
      "0 0 1 1 0 0 0 -1 0 0 6 0 "
      "0 0 1 -1 0 0 0 1 0 6 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 1 0 0 6 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 6 "
      "0 1 0 0 0 1 -1 0 0 0 6 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "P432 (P 4 3 2) 432 cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 -1 0 1 0 0 0 0 "
      "0 0 1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 -1 0 0 0 0 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "P4232 (P 42 3 2) 432 cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 -1 0 1 0 6 6 6 "
      "0 0 1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 -1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "F432 (F 4 3 2) 432 cubic 24 96 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 -1 0 1 0 0 0 0 "
      "0 0 1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 -1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "0 1 0 1 0 0 0 0 -1 0 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 6 6 "
      "0 1 0 -1 0 0 0 0 1 0 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 6 "
      "1 0 0 0 0 1 0 -1 0 0 6 6 "
      "-1 0 0 0 0 1 0 1 0 0 6 6 "
      "-1 0 0 0 0 -1 0 -1 0 0 6 6 "
      "1 0 0 0 0 -1 0 1 0 0 6 6 "
      "0 0 1 0 1 0 -1 0 0 0 6 6 "
      "0 0 1 0 -1 0 1 0 0 0 6 6 "
      "0 0 -1 0 1 0 1 0 0 0 6 6 "
      "0 0 -1 0 -1 0 -1 0 0 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 6 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 0 6 "
      "0 1 0 -1 0 0 0 0 1 6 0 6 "
      "0 -1 0 1 0 0 0 0 1 6 0 6 "
      "1 0 0 0 0 1 0 -1 0 6 0 6 "
      "-1 0 0 0 0 1 0 1 0 6 0 6 "
      "-1 0 0 0 0 -1 0 -1 0 6 0 6 "
      "1 0 0 0 0 -1 0 1 0 6 0 6 "
      "0 0 1 0 1 0 -1 0 0 6 0 6 "
      "0 0 1 0 -1 0 1 0 0 6 0 6 "
      "0 0 -1 0 1 0 1 0 0 6 0 6 "
      "0 0 -1 0 -1 0 -1 0 0 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 0 "
      "0 1 0 -1 0 0 0 0 1 6 6 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 0 "
      "1 0 0 0 0 1 0 -1 0 6 6 0 "
      "-1 0 0 0 0 1 0 1 0 6 6 0 "
      "-1 0 0 0 0 -1 0 -1 0 6 6 0 "
      "1 0 0 0 0 -1 0 1 0 6 6 0 "
      "0 0 1 0 1 0 -1 0 0 6 6 0 "
      "0 0 1 0 -1 0 1 0 0 6 6 0 "
      "0 0 -1 0 1 0 1 0 0 6 6 0 "
      "0 0 -1 0 -1 0 -1 0 0 6 6 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "F4132 (F 41 3 2) 432 cubic 24 96 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "0 1 0 1 0 0 0 0 -1 9 3 9 "
      "0 -1 0 -1 0 0 0 0 -1 3 3 3 "
      "0 1 0 -1 0 0 0 0 1 3 9 9 "
      "0 -1 0 1 0 0 0 0 1 9 9 3 "
      "1 0 0 0 0 1 0 -1 0 9 3 9 "
      "-1 0 0 0 0 1 0 1 0 9 9 3 "
      "-1 0 0 0 0 -1 0 -1 0 3 3 3 "
      "1 0 0 0 0 -1 0 1 0 3 9 9 "
      "0 0 1 0 1 0 -1 0 0 9 3 9 "
      "0 0 1 0 -1 0 1 0 0 3 9 9 "
      "0 0 -1 0 1 0 1 0 0 9 9 3 "
      "0 0 -1 0 -1 0 -1 0 0 3 3 3 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 9 9 3 "
      "0 -1 0 -1 0 0 0 0 -1 3 9 9 "
      "0 1 0 -1 0 0 0 0 1 3 3 3 "
      "0 -1 0 1 0 0 0 0 1 9 3 9 "
      "1 0 0 0 0 1 0 -1 0 9 9 3 "
      "-1 0 0 0 0 1 0 1 0 9 3 9 "
      "-1 0 0 0 0 -1 0 -1 0 3 9 9 "
      "1 0 0 0 0 -1 0 1 0 3 3 3 "
      "0 0 1 0 1 0 -1 0 0 9 9 3 "
      "0 0 1 0 -1 0 1 0 0 3 3 3 "
      "0 0 -1 0 1 0 1 0 0 9 3 9 "
      "0 0 -1 0 -1 0 -1 0 0 3 9 9 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 3 3 3 "
      "0 -1 0 -1 0 0 0 0 -1 9 3 9 "
      "0 1 0 -1 0 0 0 0 1 9 9 3 "
      "0 -1 0 1 0 0 0 0 1 3 9 9 "
      "1 0 0 0 0 1 0 -1 0 3 3 3 "
      "-1 0 0 0 0 1 0 1 0 3 9 9 "
      "-1 0 0 0 0 -1 0 -1 0 9 3 9 "
      "1 0 0 0 0 -1 0 1 0 9 9 3 "
      "0 0 1 0 1 0 -1 0 0 3 3 3 "
      "0 0 1 0 -1 0 1 0 0 9 9 3 "
      "0 0 -1 0 1 0 1 0 0 3 9 9 "
      "0 0 -1 0 -1 0 -1 0 0 9 3 9 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 3 9 9 "
      "0 -1 0 -1 0 0 0 0 -1 9 9 3 "
      "0 1 0 -1 0 0 0 0 1 9 3 9 "
      "0 -1 0 1 0 0 0 0 1 3 3 3 "
      "1 0 0 0 0 1 0 -1 0 3 9 9 "
      "-1 0 0 0 0 1 0 1 0 3 3 3 "
      "-1 0 0 0 0 -1 0 -1 0 9 9 3 "
      "1 0 0 0 0 -1 0 1 0 9 3 9 "
      "0 0 1 0 1 0 -1 0 0 3 9 9 "
      "0 0 1 0 -1 0 1 0 0 9 3 9 "
      "0 0 -1 0 1 0 1 0 0 3 3 3 "
      "0 0 -1 0 -1 0 -1 0 0 9 9 3 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "I432 (I 4 3 2) 432 cubic 24 48 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 -1 0 1 0 0 0 0 "
      "0 0 1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 -1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 -1 0 0 0 1 0 6 6 6 "
      "0 0 -1 1 0 0 0 -1 0 6 6 6 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 -1 0 0 0 -1 1 0 0 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 -1 0 1 0 6 6 6 "
      "0 0 1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 -1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "P4332 (P 43 3 2) 432 cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 3 9 9 "
      "0 -1 0 -1 0 0 0 0 -1 3 3 3 "
      "0 1 0 -1 0 0 0 0 1 9 9 3 "
      "0 -1 0 1 0 0 0 0 1 9 3 9 "
      "1 0 0 0 0 1 0 -1 0 3 9 9 "
      "-1 0 0 0 0 1 0 1 0 9 3 9 "
      "-1 0 0 0 0 -1 0 -1 0 3 3 3 "
      "1 0 0 0 0 -1 0 1 0 9 9 3 "
      "0 0 1 0 1 0 -1 0 0 3 9 9 "
      "0 0 1 0 -1 0 1 0 0 9 9 3 "
      "0 0 -1 0 1 0 1 0 0 9 3 9 "
      "0 0 -1 0 -1 0 -1 0 0 3 3 3 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "P4132 (P 41 3 2) 432 cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 9 3 3 "
      "0 -1 0 -1 0 0 0 0 -1 9 9 9 "
      "0 1 0 -1 0 0 0 0 1 3 3 9 "
      "0 -1 0 1 0 0 0 0 1 3 9 3 "
      "1 0 0 0 0 1 0 -1 0 9 3 3 "
      "-1 0 0 0 0 1 0 1 0 3 9 3 "
      "-1 0 0 0 0 -1 0 -1 0 9 9 9 "
      "1 0 0 0 0 -1 0 1 0 3 3 9 "
      "0 0 1 0 1 0 -1 0 0 9 3 3 "
      "0 0 1 0 -1 0 1 0 0 3 3 9 "
      "0 0 -1 0 1 0 1 0 0 3 9 3 "
      "0 0 -1 0 -1 0 -1 0 0 9 9 9 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "I4132 (I 41 3 2) 432 cubic 24 48 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 9 3 3 "
      "0 -1 0 -1 0 0 0 0 -1 9 9 9 "
      "0 1 0 -1 0 0 0 0 1 3 3 9 "
      "0 -1 0 1 0 0 0 0 1 3 9 3 "
      "1 0 0 0 0 1 0 -1 0 9 3 3 "
      "-1 0 0 0 0 1 0 1 0 3 9 3 "
      "-1 0 0 0 0 -1 0 -1 0 9 9 9 "
      "1 0 0 0 0 -1 0 1 0 3 3 9 "
      "0 0 1 0 1 0 -1 0 0 9 3 3 "
      "0 0 1 0 -1 0 1 0 0 3 3 9 "
      "0 0 -1 0 1 0 1 0 0 3 9 3 "
      "0 0 -1 0 -1 0 -1 0 0 9 9 9 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 0 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 0 0 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 0 "
      "0 1 0 1 0 0 0 0 -1 3 9 9 "
      "0 -1 0 -1 0 0 0 0 -1 3 3 3 "
      "0 1 0 -1 0 0 0 0 1 9 9 3 "
      "0 -1 0 1 0 0 0 0 1 9 3 9 "
      "1 0 0 0 0 1 0 -1 0 3 9 9 "
      "-1 0 0 0 0 1 0 1 0 9 3 9 "
      "-1 0 0 0 0 -1 0 -1 0 3 3 3 "
      "1 0 0 0 0 -1 0 1 0 9 9 3 "
      "0 0 1 0 1 0 -1 0 0 3 9 9 "
      "0 0 1 0 -1 0 1 0 0 9 9 3 "
      "0 0 -1 0 1 0 1 0 0 9 3 9 "
      "0 0 -1 0 -1 0 -1 0 0 3 3 3 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "P-43m (P -4 3 m) 4bar3m cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 1 0 0 0 0 "
      "1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "0 0 1 0 1 0 1 0 0 0 0 0 "
      "0 0 1 0 -1 0 -1 0 0 0 0 0 "
      "0 0 -1 0 1 0 -1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 1 0 0 0 0 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "F-43m (F -4 3 m) 4bar3m cubic 24 96 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 1 0 0 0 0 "
      "1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "0 0 1 0 1 0 1 0 0 0 0 0 "
      "0 0 1 0 -1 0 -1 0 0 0 0 0 "
      "0 0 -1 0 1 0 -1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "0 1 0 1 0 0 0 0 1 0 6 6 "
      "0 -1 0 -1 0 0 0 0 1 0 6 6 "
      "0 1 0 -1 0 0 0 0 -1 0 6 6 "
      "0 -1 0 1 0 0 0 0 -1 0 6 6 "
      "1 0 0 0 0 1 0 1 0 0 6 6 "
      "-1 0 0 0 0 1 0 -1 0 0 6 6 "
      "-1 0 0 0 0 -1 0 1 0 0 6 6 "
      "1 0 0 0 0 -1 0 -1 0 0 6 6 "
      "0 0 1 0 1 0 1 0 0 0 6 6 "
      "0 0 1 0 -1 0 -1 0 0 0 6 6 "
      "0 0 -1 0 1 0 -1 0 0 0 6 6 "
      "0 0 -1 0 -1 0 1 0 0 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 1 6 0 6 "
      "0 -1 0 -1 0 0 0 0 1 6 0 6 "
      "0 1 0 -1 0 0 0 0 -1 6 0 6 "
      "0 -1 0 1 0 0 0 0 -1 6 0 6 "
      "1 0 0 0 0 1 0 1 0 6 0 6 "
      "-1 0 0 0 0 1 0 -1 0 6 0 6 "
      "-1 0 0 0 0 -1 0 1 0 6 0 6 "
      "1 0 0 0 0 -1 0 -1 0 6 0 6 "
      "0 0 1 0 1 0 1 0 0 6 0 6 "
      "0 0 1 0 -1 0 -1 0 0 6 0 6 "
      "0 0 -1 0 1 0 -1 0 0 6 0 6 "
      "0 0 -1 0 -1 0 1 0 0 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 -1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 1 0 0 0 0 -1 6 6 0 "
      "1 0 0 0 0 1 0 1 0 6 6 0 "
      "-1 0 0 0 0 1 0 -1 0 6 6 0 "
      "-1 0 0 0 0 -1 0 1 0 6 6 0 "
      "1 0 0 0 0 -1 0 -1 0 6 6 0 "
      "0 0 1 0 1 0 1 0 0 6 6 0 "
      "0 0 1 0 -1 0 -1 0 0 6 6 0 "
      "0 0 -1 0 1 0 -1 0 0 6 6 0 "
      "0 0 -1 0 -1 0 1 0 0 6 6 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "I-43m (I -4 3 m) 4bar3m cubic 24 48 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 1 0 0 0 0 "
      "1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "0 0 1 0 1 0 1 0 0 0 0 0 "
      "0 0 1 0 -1 0 -1 0 0 0 0 0 "
      "0 0 -1 0 1 0 -1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 -1 0 0 0 1 0 6 6 6 "
      "0 0 -1 1 0 0 0 -1 0 6 6 6 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 -1 0 0 0 -1 1 0 0 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 1 0 6 6 6 "
      "1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "0 0 1 0 1 0 1 0 0 6 6 6 "
      "0 0 1 0 -1 0 -1 0 0 6 6 6 "
      "0 0 -1 0 1 0 -1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "P-43n (P -4 3 n) 4bar3m cubic 24 24 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 1 0 6 6 6 "
      "1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "0 0 1 0 1 0 1 0 0 6 6 6 "
      "0 0 1 0 -1 0 -1 0 0 6 6 6 "
      "0 0 -1 0 1 0 -1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "F-43c (F -4 3 c) 4bar3m cubic 24 96 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 1 0 6 6 6 "
      "1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "0 0 1 0 1 0 1 0 0 6 6 6 "
      "0 0 1 0 -1 0 -1 0 0 6 6 6 "
      "0 0 -1 0 1 0 -1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 1 0 0 6 6 6 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "0 1 0 1 0 0 0 0 1 6 0 0 "
      "0 -1 0 -1 0 0 0 0 1 6 0 0 "
      "0 1 0 -1 0 0 0 0 -1 6 0 0 "
      "0 -1 0 1 0 0 0 0 -1 6 0 0 "
      "1 0 0 0 0 1 0 1 0 6 0 0 "
      "-1 0 0 0 0 1 0 -1 0 6 0 0 "
      "-1 0 0 0 0 -1 0 1 0 6 0 0 "
      "1 0 0 0 0 -1 0 -1 0 6 0 0 "
      "0 0 1 0 1 0 1 0 0 6 0 0 "
      "0 0 1 0 -1 0 -1 0 0 6 0 0 "
      "0 0 -1 0 1 0 -1 0 0 6 0 0 "
      "0 0 -1 0 -1 0 1 0 0 6 0 0 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 1 0 6 0 "
      "0 -1 0 -1 0 0 0 0 1 0 6 0 "
      "0 1 0 -1 0 0 0 0 -1 0 6 0 "
      "0 -1 0 1 0 0 0 0 -1 0 6 0 "
      "1 0 0 0 0 1 0 1 0 0 6 0 "
      "-1 0 0 0 0 1 0 -1 0 0 6 0 "
      "-1 0 0 0 0 -1 0 1 0 0 6 0 "
      "1 0 0 0 0 -1 0 -1 0 0 6 0 "
      "0 0 1 0 1 0 1 0 0 0 6 0 "
      "0 0 1 0 -1 0 -1 0 0 0 6 0 "
      "0 0 -1 0 1 0 -1 0 0 0 6 0 "
      "0 0 -1 0 -1 0 1 0 0 0 6 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 6 "
      "1 0 0 0 0 1 0 1 0 0 0 6 "
      "-1 0 0 0 0 1 0 -1 0 0 0 6 "
      "-1 0 0 0 0 -1 0 1 0 0 0 6 "
      "1 0 0 0 0 -1 0 -1 0 0 0 6 "
      "0 0 1 0 1 0 1 0 0 0 0 6 "
      "0 0 1 0 -1 0 -1 0 0 0 0 6 "
      "0 0 -1 0 1 0 -1 0 0 0 0 6 "
      "0 0 -1 0 -1 0 1 0 0 0 0 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "I-43d (I -4 3 d) 4bar3m cubic 24 48 1 0 0 0 1 0 0 0 1 0 0 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 1 3 3 3 "
      "0 -1 0 -1 0 0 0 0 1 3 9 9 "
      "0 1 0 -1 0 0 0 0 -1 9 3 9 "
      "0 -1 0 1 0 0 0 0 -1 9 9 3 "
      "1 0 0 0 0 1 0 1 0 3 3 3 "
      "-1 0 0 0 0 1 0 -1 0 9 9 3 "
      "-1 0 0 0 0 -1 0 1 0 3 9 9 "
      "1 0 0 0 0 -1 0 -1 0 9 3 9 "
      "0 0 1 0 1 0 1 0 0 3 3 3 "
      "0 0 1 0 -1 0 -1 0 0 9 3 9 "
      "0 0 -1 0 1 0 -1 0 0 9 9 3 "
      "0 0 -1 0 -1 0 1 0 0 3 9 9 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 0 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 0 0 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 0 "
      "0 1 0 1 0 0 0 0 1 9 9 9 "
      "0 -1 0 -1 0 0 0 0 1 9 3 3 "
      "0 1 0 -1 0 0 0 0 -1 3 9 3 "
      "0 -1 0 1 0 0 0 0 -1 3 3 9 "
      "1 0 0 0 0 1 0 1 0 9 9 9 "
      "-1 0 0 0 0 1 0 -1 0 3 3 9 "
      "-1 0 0 0 0 -1 0 1 0 9 3 3 "
      "1 0 0 0 0 -1 0 -1 0 3 9 3 "
      "0 0 1 0 1 0 1 0 0 9 9 9 "
      "0 0 1 0 -1 0 -1 0 0 3 9 3 "
      "0 0 -1 0 1 0 -1 0 0 3 3 9 "
      "0 0 -1 0 -1 0 1 0 0 9 3 3 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Pm-3m (P 4/m -3 2/m,P m -3 m) m3barm cubic 48 48 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 -1 0 1 0 0 0 0 "
      "0 0 1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 -1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 1 0 0 0 -1 0 0 0 0 "
      "0 0 1 -1 0 0 0 1 0 0 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 0 "
      "0 1 0 0 0 1 -1 0 0 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 0 -1 0 1 0 0 0 0 "
      "1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 1 0 -1 0 0 0 0 "
      "0 0 -1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 -1 0 0 0 0 0 "
      "0 0 1 0 1 0 1 0 0 0 0 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Pn-3n (P 4/n -3 2/n P n -3 n) m3barm cubic 48 48 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 -1 0 1 0 0 0 0 "
      "0 0 1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 -1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 1 0 0 0 -1 0 6 6 6 "
      "0 0 1 -1 0 0 0 1 0 6 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 1 0 0 6 6 6 "
      "0 1 0 0 0 1 -1 0 0 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 0 -1 0 1 0 6 6 6 "
      "1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 1 0 -1 0 6 6 6 "
      "0 0 -1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 -1 0 0 6 6 6 "
      "0 0 1 0 1 0 1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Pm-3n (P 42/m -3 2/n,P m -3 n) m3barm cubic 48 48 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 -1 0 1 0 6 6 6 "
      "0 0 1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 -1 0 0 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 1 0 0 0 -1 0 0 0 0 "
      "0 0 1 -1 0 0 0 1 0 0 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 0 "
      "0 1 0 0 0 1 -1 0 0 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 0 -1 0 1 0 6 6 6 "
      "1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 1 0 -1 0 6 6 6 "
      "0 0 -1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 -1 0 0 6 6 6 "
      "0 0 1 0 1 0 1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Pn-3m (P 42/n -3 2/m,P n -3 m) m3barm cubic 48 48 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 -1 0 1 0 6 6 6 "
      "0 0 1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 -1 0 0 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 1 0 0 0 -1 0 6 6 6 "
      "0 0 1 -1 0 0 0 1 0 6 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 1 0 0 6 6 6 "
      "0 1 0 0 0 1 -1 0 0 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 0 -1 0 1 0 0 0 0 "
      "1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 1 0 -1 0 0 0 0 "
      "0 0 -1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 -1 0 0 0 0 0 "
      "0 0 1 0 1 0 1 0 0 0 0 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Fm-3m (F 4/m -3 2/m,F m -3 m) m3barm cubic 48 192 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 -1 0 1 0 0 0 0 "
      "0 0 1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 -1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 1 0 0 0 -1 0 0 0 0 "
      "0 0 1 -1 0 0 0 1 0 0 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 0 "
      "0 1 0 0 0 1 -1 0 0 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 0 -1 0 1 0 0 0 0 "
      "1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 1 0 -1 0 0 0 0 "
      "0 0 -1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 -1 0 0 0 0 0 "
      "0 0 1 0 1 0 1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "0 1 0 1 0 0 0 0 -1 0 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 6 6 "
      "0 1 0 -1 0 0 0 0 1 0 6 6 "
      "0 -1 0 1 0 0 0 0 1 0 6 6 "
      "1 0 0 0 0 1 0 -1 0 0 6 6 "
      "-1 0 0 0 0 1 0 1 0 0 6 6 "
      "-1 0 0 0 0 -1 0 -1 0 0 6 6 "
      "1 0 0 0 0 -1 0 1 0 0 6 6 "
      "0 0 1 0 1 0 -1 0 0 0 6 6 "
      "0 0 1 0 -1 0 1 0 0 0 6 6 "
      "0 0 -1 0 1 0 1 0 0 0 6 6 "
      "0 0 -1 0 -1 0 -1 0 0 0 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      "0 0 -1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 1 0 0 0 -1 0 0 6 6 "
      "0 0 1 -1 0 0 0 1 0 0 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 1 0 0 0 6 6 "
      "0 1 0 0 0 1 -1 0 0 0 6 6 "
      "0 -1 0 -1 0 0 0 0 1 0 6 6 "
      "0 1 0 1 0 0 0 0 1 0 6 6 "
      "0 -1 0 1 0 0 0 0 -1 0 6 6 "
      "0 1 0 -1 0 0 0 0 -1 0 6 6 "
      "-1 0 0 0 0 -1 0 1 0 0 6 6 "
      "1 0 0 0 0 -1 0 -1 0 0 6 6 "
      "1 0 0 0 0 1 0 1 0 0 6 6 "
      "-1 0 0 0 0 1 0 -1 0 0 6 6 "
      "0 0 -1 0 -1 0 1 0 0 0 6 6 "
      "0 0 -1 0 1 0 -1 0 0 0 6 6 "
      "0 0 1 0 -1 0 -1 0 0 0 6 6 "
      "0 0 1 0 1 0 1 0 0 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 6 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 0 6 "
      "0 1 0 -1 0 0 0 0 1 6 0 6 "
      "0 -1 0 1 0 0 0 0 1 6 0 6 "
      "1 0 0 0 0 1 0 -1 0 6 0 6 "
      "-1 0 0 0 0 1 0 1 0 6 0 6 "
      "-1 0 0 0 0 -1 0 -1 0 6 0 6 "
      "1 0 0 0 0 -1 0 1 0 6 0 6 "
      "0 0 1 0 1 0 -1 0 0 6 0 6 "
      "0 0 1 0 -1 0 1 0 0 6 0 6 "
      "0 0 -1 0 1 0 1 0 0 6 0 6 "
      "0 0 -1 0 -1 0 -1 0 0 6 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 1 0 0 0 -1 0 6 0 6 "
      "0 0 1 -1 0 0 0 1 0 6 0 6 "
      "0 -1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 1 0 0 6 0 6 "
      "0 1 0 0 0 1 -1 0 0 6 0 6 "
      "0 -1 0 -1 0 0 0 0 1 6 0 6 "
      "0 1 0 1 0 0 0 0 1 6 0 6 "
      "0 -1 0 1 0 0 0 0 -1 6 0 6 "
      "0 1 0 -1 0 0 0 0 -1 6 0 6 "
      "-1 0 0 0 0 -1 0 1 0 6 0 6 "
      "1 0 0 0 0 -1 0 -1 0 6 0 6 "
      "1 0 0 0 0 1 0 1 0 6 0 6 "
      "-1 0 0 0 0 1 0 -1 0 6 0 6 "
      "0 0 -1 0 -1 0 1 0 0 6 0 6 "
      "0 0 -1 0 1 0 -1 0 0 6 0 6 "
      "0 0 1 0 -1 0 -1 0 0 6 0 6 "
      "0 0 1 0 1 0 1 0 0 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 0 "
      "0 1 0 -1 0 0 0 0 1 6 6 0 "
      "0 -1 0 1 0 0 0 0 1 6 6 0 "
      "1 0 0 0 0 1 0 -1 0 6 6 0 "
      "-1 0 0 0 0 1 0 1 0 6 6 0 "
      "-1 0 0 0 0 -1 0 -1 0 6 6 0 "
      "1 0 0 0 0 -1 0 1 0 6 6 0 "
      "0 0 1 0 1 0 -1 0 0 6 6 0 "
      "0 0 1 0 -1 0 1 0 0 6 6 0 "
      "0 0 -1 0 1 0 1 0 0 6 6 0 "
      "0 0 -1 0 -1 0 -1 0 0 6 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 1 0 0 0 -1 0 6 6 0 "
      "0 0 1 -1 0 0 0 1 0 6 6 0 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 1 0 0 6 6 0 "
      "0 1 0 0 0 1 -1 0 0 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      "0 -1 0 1 0 0 0 0 -1 6 6 0 "
      "0 1 0 -1 0 0 0 0 -1 6 6 0 "
      "-1 0 0 0 0 -1 0 1 0 6 6 0 "
      "1 0 0 0 0 -1 0 -1 0 6 6 0 "
      "1 0 0 0 0 1 0 1 0 6 6 0 "
      "-1 0 0 0 0 1 0 -1 0 6 6 0 "
      "0 0 -1 0 -1 0 1 0 0 6 6 0 "
      "0 0 -1 0 1 0 -1 0 0 6 6 0 "
      "0 0 1 0 -1 0 -1 0 0 6 6 0 "
      "0 0 1 0 1 0 1 0 0 6 6 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Fm-3c (F 4/m -3 2/c,F m -3 c) m3barm cubic 48 192 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 -1 0 1 0 6 6 6 "
      "0 0 1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 -1 0 0 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 1 0 0 0 -1 0 0 0 0 "
      "0 0 1 -1 0 0 0 1 0 0 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 0 "
      "0 1 0 0 0 1 -1 0 0 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 0 -1 0 1 0 6 6 6 "
      "1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 1 0 -1 0 6 6 6 "
      "0 0 -1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 -1 0 0 6 6 6 "
      "0 0 1 0 1 0 1 0 0 6 6 6 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "0 1 0 1 0 0 0 0 -1 6 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 6 0 0 "
      "0 1 0 -1 0 0 0 0 1 6 0 0 "
      "0 -1 0 1 0 0 0 0 1 6 0 0 "
      "1 0 0 0 0 1 0 -1 0 6 0 0 "
      "-1 0 0 0 0 1 0 1 0 6 0 0 "
      "-1 0 0 0 0 -1 0 -1 0 6 0 0 "
      "1 0 0 0 0 -1 0 1 0 6 0 0 "
      "0 0 1 0 1 0 -1 0 0 6 0 0 "
      "0 0 1 0 -1 0 1 0 0 6 0 0 "
      "0 0 -1 0 1 0 1 0 0 6 0 0 "
      "0 0 -1 0 -1 0 -1 0 0 6 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 0 6 6 "
      "0 0 -1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 1 0 0 0 -1 0 0 6 6 "
      "0 0 1 -1 0 0 0 1 0 0 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 1 0 0 0 6 6 "
      "0 1 0 0 0 1 -1 0 0 0 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 0 0 "
      "0 1 0 1 0 0 0 0 1 6 0 0 "
      "0 -1 0 1 0 0 0 0 -1 6 0 0 "
      "0 1 0 -1 0 0 0 0 -1 6 0 0 "
      "-1 0 0 0 0 -1 0 1 0 6 0 0 "
      "1 0 0 0 0 -1 0 -1 0 6 0 0 "
      "1 0 0 0 0 1 0 1 0 6 0 0 "
      "-1 0 0 0 0 1 0 -1 0 6 0 0 "
      "0 0 -1 0 -1 0 1 0 0 6 0 0 "
      "0 0 -1 0 1 0 -1 0 0 6 0 0 "
      "0 0 1 0 -1 0 -1 0 0 6 0 0 "
      "0 0 1 0 1 0 1 0 0 6 0 0 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 0 6 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 6 0 "
      "0 1 0 -1 0 0 0 0 1 0 6 0 "
      "0 -1 0 1 0 0 0 0 1 0 6 0 "
      "1 0 0 0 0 1 0 -1 0 0 6 0 "
      "-1 0 0 0 0 1 0 1 0 0 6 0 "
      "-1 0 0 0 0 -1 0 -1 0 0 6 0 "
      "1 0 0 0 0 -1 0 1 0 0 6 0 "
      "0 0 1 0 1 0 -1 0 0 0 6 0 "
      "0 0 1 0 -1 0 1 0 0 0 6 0 "
      "0 0 -1 0 1 0 1 0 0 0 6 0 "
      "0 0 -1 0 -1 0 -1 0 0 0 6 0 "
      "-1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 1 6 0 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 1 0 0 0 -1 0 6 0 6 "
      "0 0 1 -1 0 0 0 1 0 6 0 6 "
      "0 -1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 1 0 0 6 0 6 "
      "0 1 0 0 0 1 -1 0 0 6 0 6 "
      "0 -1 0 -1 0 0 0 0 1 0 6 0 "
      "0 1 0 1 0 0 0 0 1 0 6 0 "
      "0 -1 0 1 0 0 0 0 -1 0 6 0 "
      "0 1 0 -1 0 0 0 0 -1 0 6 0 "
      "-1 0 0 0 0 -1 0 1 0 0 6 0 "
      "1 0 0 0 0 -1 0 -1 0 0 6 0 "
      "1 0 0 0 0 1 0 1 0 0 6 0 "
      "-1 0 0 0 0 1 0 -1 0 0 6 0 "
      "0 0 -1 0 -1 0 1 0 0 0 6 0 "
      "0 0 -1 0 1 0 -1 0 0 0 6 0 "
      "0 0 1 0 -1 0 -1 0 0 0 6 0 "
      "0 0 1 0 1 0 1 0 0 0 6 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 6 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 6 "
      "0 1 0 -1 0 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 1 0 0 6 "
      "1 0 0 0 0 1 0 -1 0 0 0 6 "
      "-1 0 0 0 0 1 0 1 0 0 0 6 "
      "-1 0 0 0 0 -1 0 -1 0 0 0 6 "
      "1 0 0 0 0 -1 0 1 0 0 0 6 "
      "0 0 1 0 1 0 -1 0 0 0 0 6 "
      "0 0 1 0 -1 0 1 0 0 0 0 6 "
      "0 0 -1 0 1 0 1 0 0 0 0 6 "
      "0 0 -1 0 -1 0 -1 0 0 0 0 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 1 0 0 0 -1 0 6 6 0 "
      "0 0 1 -1 0 0 0 1 0 6 6 0 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 1 0 0 6 6 0 "
      "0 1 0 0 0 1 -1 0 0 6 6 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 6 "
      "0 1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 0 0 0 0 -1 0 1 0 0 0 6 "
      "1 0 0 0 0 -1 0 -1 0 0 0 6 "
      "1 0 0 0 0 1 0 1 0 0 0 6 "
      "-1 0 0 0 0 1 0 -1 0 0 0 6 "
      "0 0 -1 0 -1 0 1 0 0 0 0 6 "
      "0 0 -1 0 1 0 -1 0 0 0 0 6 "
      "0 0 1 0 -1 0 -1 0 0 0 0 6 "
      "0 0 1 0 1 0 1 0 0 0 0 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Fd-3m (F 41/d -3 2/m,F d -3 m) m3barm cubic 48 192 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "0 1 0 1 0 0 0 0 -1 9 3 9 "
      "0 -1 0 -1 0 0 0 0 -1 3 3 3 "
      "0 1 0 -1 0 0 0 0 1 3 9 9 "
      "0 -1 0 1 0 0 0 0 1 9 9 3 "
      "1 0 0 0 0 1 0 -1 0 9 3 9 "
      "-1 0 0 0 0 1 0 1 0 9 9 3 "
      "-1 0 0 0 0 -1 0 -1 0 3 3 3 "
      "1 0 0 0 0 -1 0 1 0 3 9 9 "
      "0 0 1 0 1 0 -1 0 0 9 3 9 "
      "0 0 1 0 -1 0 1 0 0 3 9 9 "
      "0 0 -1 0 1 0 1 0 0 9 9 3 "
      "0 0 -1 0 -1 0 -1 0 0 3 3 3 "
      "-1 0 0 0 -1 0 0 0 -1 3 3 3 "
      "1 0 0 0 1 0 0 0 -1 3 9 9 "
      "1 0 0 0 -1 0 0 0 1 9 9 3 "
      "-1 0 0 0 1 0 0 0 1 9 3 9 "
      "0 0 -1 -1 0 0 0 -1 0 3 3 3 "
      "0 0 -1 1 0 0 0 1 0 9 3 9 "
      "0 0 1 1 0 0 0 -1 0 3 9 9 "
      "0 0 1 -1 0 0 0 1 0 9 9 3 "
      "0 -1 0 0 0 -1 -1 0 0 3 3 3 "
      "0 1 0 0 0 -1 1 0 0 9 9 3 "
      "0 -1 0 0 0 1 1 0 0 9 3 9 "
      "0 1 0 0 0 1 -1 0 0 3 9 9 "
      "0 -1 0 -1 0 0 0 0 1 6 0 6 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 0 "
      "-1 0 0 0 0 -1 0 1 0 6 0 6 "
      "1 0 0 0 0 -1 0 -1 0 6 6 0 "
      "1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 1 0 -1 0 0 6 6 "
      "0 0 -1 0 -1 0 1 0 0 6 0 6 "
      "0 0 -1 0 1 0 -1 0 0 0 6 6 "
      "0 0 1 0 -1 0 -1 0 0 6 6 0 "
      "0 0 1 0 1 0 1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 9 9 3 "
      "0 -1 0 -1 0 0 0 0 -1 3 9 9 "
      "0 1 0 -1 0 0 0 0 1 3 3 3 "
      "0 -1 0 1 0 0 0 0 1 9 3 9 "
      "1 0 0 0 0 1 0 -1 0 9 9 3 "
      "-1 0 0 0 0 1 0 1 0 9 3 9 "
      "-1 0 0 0 0 -1 0 -1 0 3 9 9 "
      "1 0 0 0 0 -1 0 1 0 3 3 3 "
      "0 0 1 0 1 0 -1 0 0 9 9 3 "
      "0 0 1 0 -1 0 1 0 0 3 3 3 "
      "0 0 -1 0 1 0 1 0 0 9 3 9 "
      "0 0 -1 0 -1 0 -1 0 0 3 9 9 "
      "-1 0 0 0 -1 0 0 0 -1 3 9 9 "
      "1 0 0 0 1 0 0 0 -1 3 3 3 "
      "1 0 0 0 -1 0 0 0 1 9 3 9 "
      "-1 0 0 0 1 0 0 0 1 9 9 3 "
      "0 0 -1 -1 0 0 0 -1 0 3 9 9 "
      "0 0 -1 1 0 0 0 1 0 9 9 3 "
      "0 0 1 1 0 0 0 -1 0 3 3 3 "
      "0 0 1 -1 0 0 0 1 0 9 3 9 "
      "0 -1 0 0 0 -1 -1 0 0 3 9 9 "
      "0 1 0 0 0 -1 1 0 0 9 3 9 "
      "0 -1 0 0 0 1 1 0 0 9 9 3 "
      "0 1 0 0 0 1 -1 0 0 3 3 3 "
      "0 -1 0 -1 0 0 0 0 1 6 6 0 "
      "0 1 0 1 0 0 0 0 1 0 6 6 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 6 0 6 "
      "-1 0 0 0 0 -1 0 1 0 6 6 0 "
      "1 0 0 0 0 -1 0 -1 0 6 0 6 "
      "1 0 0 0 0 1 0 1 0 0 6 6 "
      "-1 0 0 0 0 1 0 -1 0 0 0 0 "
      "0 0 -1 0 -1 0 1 0 0 6 6 0 "
      "0 0 -1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 -1 0 0 6 0 6 "
      "0 0 1 0 1 0 1 0 0 0 6 6 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 3 3 3 "
      "0 -1 0 -1 0 0 0 0 -1 9 3 9 "
      "0 1 0 -1 0 0 0 0 1 9 9 3 "
      "0 -1 0 1 0 0 0 0 1 3 9 9 "
      "1 0 0 0 0 1 0 -1 0 3 3 3 "
      "-1 0 0 0 0 1 0 1 0 3 9 9 "
      "-1 0 0 0 0 -1 0 -1 0 9 3 9 "
      "1 0 0 0 0 -1 0 1 0 9 9 3 "
      "0 0 1 0 1 0 -1 0 0 3 3 3 "
      "0 0 1 0 -1 0 1 0 0 9 9 3 "
      "0 0 -1 0 1 0 1 0 0 3 9 9 "
      "0 0 -1 0 -1 0 -1 0 0 9 3 9 "
      "-1 0 0 0 -1 0 0 0 -1 9 3 9 "
      "1 0 0 0 1 0 0 0 -1 9 9 3 "
      "1 0 0 0 -1 0 0 0 1 3 9 9 "
      "-1 0 0 0 1 0 0 0 1 3 3 3 "
      "0 0 -1 -1 0 0 0 -1 0 9 3 9 "
      "0 0 -1 1 0 0 0 1 0 3 3 3 "
      "0 0 1 1 0 0 0 -1 0 9 9 3 "
      "0 0 1 -1 0 0 0 1 0 3 9 9 "
      "0 -1 0 0 0 -1 -1 0 0 9 3 9 "
      "0 1 0 0 0 -1 1 0 0 3 9 9 "
      "0 -1 0 0 0 1 1 0 0 3 3 3 "
      "0 1 0 0 0 1 -1 0 0 9 9 3 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 6 0 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 0 "
      "0 1 0 -1 0 0 0 0 -1 0 6 6 "
      "-1 0 0 0 0 -1 0 1 0 0 0 0 "
      "1 0 0 0 0 -1 0 -1 0 0 6 6 "
      "1 0 0 0 0 1 0 1 0 6 0 6 "
      "-1 0 0 0 0 1 0 -1 0 6 6 0 "
      "0 0 -1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 -1 0 0 6 6 0 "
      "0 0 1 0 -1 0 -1 0 0 0 6 6 "
      "0 0 1 0 1 0 1 0 0 6 0 6 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 3 9 9 "
      "0 -1 0 -1 0 0 0 0 -1 9 9 3 "
      "0 1 0 -1 0 0 0 0 1 9 3 9 "
      "0 -1 0 1 0 0 0 0 1 3 3 3 "
      "1 0 0 0 0 1 0 -1 0 3 9 9 "
      "-1 0 0 0 0 1 0 1 0 3 3 3 "
      "-1 0 0 0 0 -1 0 -1 0 9 9 3 "
      "1 0 0 0 0 -1 0 1 0 9 3 9 "
      "0 0 1 0 1 0 -1 0 0 3 9 9 "
      "0 0 1 0 -1 0 1 0 0 9 3 9 "
      "0 0 -1 0 1 0 1 0 0 3 3 3 "
      "0 0 -1 0 -1 0 -1 0 0 9 9 3 "
      "-1 0 0 0 -1 0 0 0 -1 9 9 3 "
      "1 0 0 0 1 0 0 0 -1 9 3 9 "
      "1 0 0 0 -1 0 0 0 1 3 3 3 "
      "-1 0 0 0 1 0 0 0 1 3 9 9 "
      "0 0 -1 -1 0 0 0 -1 0 9 9 3 "
      "0 0 -1 1 0 0 0 1 0 3 9 9 "
      "0 0 1 1 0 0 0 -1 0 9 3 9 "
      "0 0 1 -1 0 0 0 1 0 3 3 3 "
      "0 -1 0 0 0 -1 -1 0 0 9 9 3 "
      "0 1 0 0 0 -1 1 0 0 3 3 3 "
      "0 -1 0 0 0 1 1 0 0 3 9 9 "
      "0 1 0 0 0 1 -1 0 0 9 3 9 "
      "0 -1 0 -1 0 0 0 0 1 0 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 0 "
      "0 -1 0 1 0 0 0 0 -1 6 0 6 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 0 -1 0 1 0 0 6 6 "
      "1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 1 0 1 0 6 6 0 "
      "-1 0 0 0 0 1 0 -1 0 6 0 6 "
      "0 0 -1 0 -1 0 1 0 0 0 6 6 "
      "0 0 -1 0 1 0 -1 0 0 6 0 6 "
      "0 0 1 0 -1 0 -1 0 0 0 0 0 "
      "0 0 1 0 1 0 1 0 0 6 6 0 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Fd-3c (F 41/d -3 2/c,F d -3 c) m3barm cubic 48 192 1 0 0 0 1 0 0 0 1 0 "
      "0 0 "
      "-1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 0 "
      "1 0 0 0 -1 0 0 0 -1 6 0 6 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 0 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 6 "
      "0 0 -1 1 0 0 0 -1 0 6 6 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 6 6 0 "
      "0 1 0 0 0 -1 -1 0 0 6 0 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 6 "
      "0 1 0 1 0 0 0 0 -1 9 3 9 "
      "0 -1 0 -1 0 0 0 0 -1 3 3 3 "
      "0 1 0 -1 0 0 0 0 1 3 9 9 "
      "0 -1 0 1 0 0 0 0 1 9 9 3 "
      "1 0 0 0 0 1 0 -1 0 9 3 9 "
      "-1 0 0 0 0 1 0 1 0 9 9 3 "
      "-1 0 0 0 0 -1 0 -1 0 3 3 3 "
      "1 0 0 0 0 -1 0 1 0 3 9 9 "
      "0 0 1 0 1 0 -1 0 0 9 3 9 "
      "0 0 1 0 -1 0 1 0 0 3 9 9 "
      "0 0 -1 0 1 0 1 0 0 9 9 3 "
      "0 0 -1 0 -1 0 -1 0 0 3 3 3 "
      "-1 0 0 0 -1 0 0 0 -1 9 9 9 "
      "1 0 0 0 1 0 0 0 -1 9 3 3 "
      "1 0 0 0 -1 0 0 0 1 3 3 9 "
      "-1 0 0 0 1 0 0 0 1 3 9 3 "
      "0 0 -1 -1 0 0 0 -1 0 9 9 9 "
      "0 0 -1 1 0 0 0 1 0 3 9 3 "
      "0 0 1 1 0 0 0 -1 0 9 3 3 "
      "0 0 1 -1 0 0 0 1 0 3 3 9 "
      "0 -1 0 0 0 -1 -1 0 0 9 9 9 "
      "0 1 0 0 0 -1 1 0 0 3 3 9 "
      "0 -1 0 0 0 1 1 0 0 3 9 3 "
      "0 1 0 0 0 1 -1 0 0 9 3 3 "
      "0 -1 0 -1 0 0 0 0 1 0 6 0 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 6 "
      "-1 0 0 0 0 -1 0 1 0 0 6 0 "
      "1 0 0 0 0 -1 0 -1 0 0 0 6 "
      "1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 1 0 -1 0 6 0 0 "
      "0 0 -1 0 -1 0 1 0 0 0 6 0 "
      "0 0 -1 0 1 0 -1 0 0 6 0 0 "
      "0 0 1 0 -1 0 -1 0 0 0 0 6 "
      "0 0 1 0 1 0 1 0 0 6 6 6 "
      "1 0 0 0 1 0 0 0 1 0 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 6 6 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 6 0 6 "
      "0 1 0 0 0 1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 9 9 3 "
      "0 -1 0 -1 0 0 0 0 -1 3 9 9 "
      "0 1 0 -1 0 0 0 0 1 3 3 3 "
      "0 -1 0 1 0 0 0 0 1 9 3 9 "
      "1 0 0 0 0 1 0 -1 0 9 9 3 "
      "-1 0 0 0 0 1 0 1 0 9 3 9 "
      "-1 0 0 0 0 -1 0 -1 0 3 9 9 "
      "1 0 0 0 0 -1 0 1 0 3 3 3 "
      "0 0 1 0 1 0 -1 0 0 9 9 3 "
      "0 0 1 0 -1 0 1 0 0 3 3 3 "
      "0 0 -1 0 1 0 1 0 0 9 3 9 "
      "0 0 -1 0 -1 0 -1 0 0 3 9 9 "
      "-1 0 0 0 -1 0 0 0 -1 9 3 3 "
      "1 0 0 0 1 0 0 0 -1 9 9 9 "
      "1 0 0 0 -1 0 0 0 1 3 9 3 "
      "-1 0 0 0 1 0 0 0 1 3 3 9 "
      "0 0 -1 -1 0 0 0 -1 0 9 3 3 "
      "0 0 -1 1 0 0 0 1 0 3 3 9 "
      "0 0 1 1 0 0 0 -1 0 9 9 9 "
      "0 0 1 -1 0 0 0 1 0 3 9 3 "
      "0 -1 0 0 0 -1 -1 0 0 9 3 3 "
      "0 1 0 0 0 -1 1 0 0 3 9 3 "
      "0 -1 0 0 0 1 1 0 0 3 3 9 "
      "0 1 0 0 0 1 -1 0 0 9 9 9 "
      "0 -1 0 -1 0 0 0 0 1 0 0 6 "
      "0 1 0 1 0 0 0 0 1 6 0 0 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 0 6 0 "
      "-1 0 0 0 0 -1 0 1 0 0 0 6 "
      "1 0 0 0 0 -1 0 -1 0 0 6 0 "
      "1 0 0 0 0 1 0 1 0 6 0 0 "
      "-1 0 0 0 0 1 0 -1 0 6 6 6 "
      "0 0 -1 0 -1 0 1 0 0 0 0 6 "
      "0 0 -1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 -1 0 0 0 6 0 "
      "0 0 1 0 1 0 1 0 0 6 0 0 "
      "1 0 0 0 1 0 0 0 1 6 0 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 0 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 6 0 6 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 6 6 0 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 6 0 6 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 6 6 0 "
      "0 1 0 1 0 0 0 0 -1 3 3 3 "
      "0 -1 0 -1 0 0 0 0 -1 9 3 9 "
      "0 1 0 -1 0 0 0 0 1 9 9 3 "
      "0 -1 0 1 0 0 0 0 1 3 9 9 "
      "1 0 0 0 0 1 0 -1 0 3 3 3 "
      "-1 0 0 0 0 1 0 1 0 3 9 9 "
      "-1 0 0 0 0 -1 0 -1 0 9 3 9 "
      "1 0 0 0 0 -1 0 1 0 9 9 3 "
      "0 0 1 0 1 0 -1 0 0 3 3 3 "
      "0 0 1 0 -1 0 1 0 0 9 9 3 "
      "0 0 -1 0 1 0 1 0 0 3 9 9 "
      "0 0 -1 0 -1 0 -1 0 0 9 3 9 "
      "-1 0 0 0 -1 0 0 0 -1 3 9 3 "
      "1 0 0 0 1 0 0 0 -1 3 3 9 "
      "1 0 0 0 -1 0 0 0 1 9 3 3 "
      "-1 0 0 0 1 0 0 0 1 9 9 9 "
      "0 0 -1 -1 0 0 0 -1 0 3 9 3 "
      "0 0 -1 1 0 0 0 1 0 9 9 9 "
      "0 0 1 1 0 0 0 -1 0 3 3 9 "
      "0 0 1 -1 0 0 0 1 0 9 3 3 "
      "0 -1 0 0 0 -1 -1 0 0 3 9 3 "
      "0 1 0 0 0 -1 1 0 0 9 3 3 "
      "0 -1 0 0 0 1 1 0 0 9 9 9 "
      "0 1 0 0 0 1 -1 0 0 3 3 9 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 0 6 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 6 "
      "0 1 0 -1 0 0 0 0 -1 6 0 0 "
      "-1 0 0 0 0 -1 0 1 0 6 6 6 "
      "1 0 0 0 0 -1 0 -1 0 6 0 0 "
      "1 0 0 0 0 1 0 1 0 0 6 0 "
      "-1 0 0 0 0 1 0 -1 0 0 0 6 "
      "0 0 -1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 -1 0 0 0 0 6 "
      "0 0 1 0 -1 0 -1 0 0 6 0 0 "
      "0 0 1 0 1 0 1 0 0 0 6 0 "
      "1 0 0 0 1 0 0 0 1 6 6 0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 6 6 "
      "0 0 1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 -1 0 0 0 -1 0 0 6 6 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 6 6 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 6 6 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 3 9 9 "
      "0 -1 0 -1 0 0 0 0 -1 9 9 3 "
      "0 1 0 -1 0 0 0 0 1 9 3 9 "
      "0 -1 0 1 0 0 0 0 1 3 3 3 "
      "1 0 0 0 0 1 0 -1 0 3 9 9 "
      "-1 0 0 0 0 1 0 1 0 3 3 3 "
      "-1 0 0 0 0 -1 0 -1 0 9 9 3 "
      "1 0 0 0 0 -1 0 1 0 9 3 9 "
      "0 0 1 0 1 0 -1 0 0 3 9 9 "
      "0 0 1 0 -1 0 1 0 0 9 3 9 "
      "0 0 -1 0 1 0 1 0 0 3 3 3 "
      "0 0 -1 0 -1 0 -1 0 0 9 9 3 "
      "-1 0 0 0 -1 0 0 0 -1 3 3 9 "
      "1 0 0 0 1 0 0 0 -1 3 9 3 "
      "1 0 0 0 -1 0 0 0 1 9 9 9 "
      "-1 0 0 0 1 0 0 0 1 9 3 3 "
      "0 0 -1 -1 0 0 0 -1 0 3 3 9 "
      "0 0 -1 1 0 0 0 1 0 9 3 3 "
      "0 0 1 1 0 0 0 -1 0 3 9 3 "
      "0 0 1 -1 0 0 0 1 0 9 9 9 "
      "0 -1 0 0 0 -1 -1 0 0 3 3 9 "
      "0 1 0 0 0 -1 1 0 0 9 9 9 "
      "0 -1 0 0 0 1 1 0 0 9 3 3 "
      "0 1 0 0 0 1 -1 0 0 3 9 3 "
      "0 -1 0 -1 0 0 0 0 1 6 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 6 "
      "0 -1 0 1 0 0 0 0 -1 0 6 0 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 0 -1 0 1 0 6 0 0 "
      "1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 1 0 1 0 0 0 6 "
      "-1 0 0 0 0 1 0 -1 0 0 6 0 "
      "0 0 -1 0 -1 0 1 0 0 6 0 0 "
      "0 0 -1 0 1 0 -1 0 0 0 6 0 "
      "0 0 1 0 -1 0 -1 0 0 6 6 6 "
      "0 0 1 0 1 0 1 0 0 0 0 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Im-3m (I 4/m -3 2/m,I m -3 m) m3barm cubic 48 96 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 -1 0 0 0 1 0 0 0 0 "
      "0 0 -1 1 0 0 0 -1 0 0 0 0 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 -1 0 0 0 -1 1 0 0 0 0 0 "
      "0 1 0 1 0 0 0 0 -1 0 0 0 "
      "0 -1 0 -1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 1 0 0 0 "
      "1 0 0 0 0 1 0 -1 0 0 0 0 "
      "-1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 -1 0 1 0 0 0 0 "
      "0 0 1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 1 0 0 0 0 0 "
      "0 0 -1 0 -1 0 -1 0 0 0 0 0 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 0 0 0 "
      "1 0 0 0 -1 0 0 0 1 0 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 1 0 0 0 -1 0 0 0 0 "
      "0 0 1 -1 0 0 0 1 0 0 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 0 "
      "0 1 0 0 0 1 -1 0 0 0 0 0 "
      "0 -1 0 -1 0 0 0 0 1 0 0 0 "
      "0 1 0 1 0 0 0 0 1 0 0 0 "
      "0 -1 0 1 0 0 0 0 -1 0 0 0 "
      "0 1 0 -1 0 0 0 0 -1 0 0 0 "
      "-1 0 0 0 0 -1 0 1 0 0 0 0 "
      "1 0 0 0 0 -1 0 -1 0 0 0 0 "
      "1 0 0 0 0 1 0 1 0 0 0 0 "
      "-1 0 0 0 0 1 0 -1 0 0 0 0 "
      "0 0 -1 0 -1 0 1 0 0 0 0 0 "
      "0 0 -1 0 1 0 -1 0 0 0 0 0 "
      "0 0 1 0 -1 0 -1 0 0 0 0 0 "
      "0 0 1 0 1 0 1 0 0 0 0 0 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 -1 0 0 0 1 0 6 6 6 "
      "0 0 -1 1 0 0 0 -1 0 6 6 6 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 -1 0 0 0 -1 1 0 0 6 6 6 "
      "0 1 0 1 0 0 0 0 -1 6 6 6 "
      "0 -1 0 -1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 1 6 6 6 "
      "1 0 0 0 0 1 0 -1 0 6 6 6 "
      "-1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 -1 0 1 0 6 6 6 "
      "0 0 1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 1 0 0 6 6 6 "
      "0 0 -1 0 -1 0 -1 0 0 6 6 6 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 6 6 6 "
      "1 0 0 0 -1 0 0 0 1 6 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 1 0 0 0 -1 0 6 6 6 "
      "0 0 1 -1 0 0 0 1 0 6 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 1 0 0 6 6 6 "
      "0 1 0 0 0 1 -1 0 0 6 6 6 "
      "0 -1 0 -1 0 0 0 0 1 6 6 6 "
      "0 1 0 1 0 0 0 0 1 6 6 6 "
      "0 -1 0 1 0 0 0 0 -1 6 6 6 "
      "0 1 0 -1 0 0 0 0 -1 6 6 6 "
      "-1 0 0 0 0 -1 0 1 0 6 6 6 "
      "1 0 0 0 0 -1 0 -1 0 6 6 6 "
      "1 0 0 0 0 1 0 1 0 6 6 6 "
      "-1 0 0 0 0 1 0 -1 0 6 6 6 "
      "0 0 -1 0 -1 0 1 0 0 6 6 6 "
      "0 0 -1 0 1 0 -1 0 0 6 6 6 "
      "0 0 1 0 -1 0 -1 0 0 6 6 6 "
      "0 0 1 0 1 0 1 0 0 6 6 6 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      "Ia-3d (I 41/a -3 2/d,I a -3 d) m3barm cubic 48 96 1 0 0 0 1 0 0 0 1 0 0 "
      "0 "
      "-1 0 0 0 -1 0 0 0 1 6 0 6 "
      "-1 0 0 0 1 0 0 0 -1 0 6 6 "
      "1 0 0 0 -1 0 0 0 -1 6 6 0 "
      "0 0 1 1 0 0 0 1 0 0 0 0 "
      "0 0 1 -1 0 0 0 -1 0 6 6 0 "
      "0 0 -1 -1 0 0 0 1 0 6 0 6 "
      "0 0 -1 1 0 0 0 -1 0 0 6 6 "
      "0 1 0 0 0 1 1 0 0 0 0 0 "
      "0 -1 0 0 0 1 -1 0 0 0 6 6 "
      "0 1 0 0 0 -1 -1 0 0 6 6 0 "
      "0 -1 0 0 0 -1 1 0 0 6 0 6 "
      "0 1 0 1 0 0 0 0 -1 9 3 3 "
      "0 -1 0 -1 0 0 0 0 -1 9 9 9 "
      "0 1 0 -1 0 0 0 0 1 3 3 9 "
      "0 -1 0 1 0 0 0 0 1 3 9 3 "
      "1 0 0 0 0 1 0 -1 0 9 3 3 "
      "-1 0 0 0 0 1 0 1 0 3 9 3 "
      "-1 0 0 0 0 -1 0 -1 0 9 9 9 "
      "1 0 0 0 0 -1 0 1 0 3 3 9 "
      "0 0 1 0 1 0 -1 0 0 9 3 3 "
      "0 0 1 0 -1 0 1 0 0 3 3 9 "
      "0 0 -1 0 1 0 1 0 0 3 9 3 "
      "0 0 -1 0 -1 0 -1 0 0 9 9 9 "
      "-1 0 0 0 -1 0 0 0 -1 0 0 0 "
      "1 0 0 0 1 0 0 0 -1 6 0 6 "
      "1 0 0 0 -1 0 0 0 1 0 6 6 "
      "-1 0 0 0 1 0 0 0 1 6 6 0 "
      "0 0 -1 -1 0 0 0 -1 0 0 0 0 "
      "0 0 -1 1 0 0 0 1 0 6 6 0 "
      "0 0 1 1 0 0 0 -1 0 6 0 6 "
      "0 0 1 -1 0 0 0 1 0 0 6 6 "
      "0 -1 0 0 0 -1 -1 0 0 0 0 0 "
      "0 1 0 0 0 -1 1 0 0 0 6 6 "
      "0 -1 0 0 0 1 1 0 0 6 6 0 "
      "0 1 0 0 0 1 -1 0 0 6 0 6 "
      "0 -1 0 -1 0 0 0 0 1 3 9 9 "
      "0 1 0 1 0 0 0 0 1 3 3 3 "
      "0 -1 0 1 0 0 0 0 -1 9 9 3 "
      "0 1 0 -1 0 0 0 0 -1 9 3 9 "
      "-1 0 0 0 0 -1 0 1 0 3 9 9 "
      "1 0 0 0 0 -1 0 -1 0 9 3 9 "
      "1 0 0 0 0 1 0 1 0 3 3 3 "
      "-1 0 0 0 0 1 0 -1 0 9 9 3 "
      "0 0 -1 0 -1 0 1 0 0 3 9 9 "
      "0 0 -1 0 1 0 -1 0 0 9 9 3 "
      "0 0 1 0 -1 0 -1 0 0 9 3 9 "
      "0 0 1 0 1 0 1 0 0 3 3 3 "
      "1 0 0 0 1 0 0 0 1 6 6 6 "
      "-1 0 0 0 -1 0 0 0 1 0 6 0 "
      "-1 0 0 0 1 0 0 0 -1 6 0 0 "
      "1 0 0 0 -1 0 0 0 -1 0 0 6 "
      "0 0 1 1 0 0 0 1 0 6 6 6 "
      "0 0 1 -1 0 0 0 -1 0 0 0 6 "
      "0 0 -1 -1 0 0 0 1 0 0 6 0 "
      "0 0 -1 1 0 0 0 -1 0 6 0 0 "
      "0 1 0 0 0 1 1 0 0 6 6 6 "
      "0 -1 0 0 0 1 -1 0 0 6 0 0 "
      "0 1 0 0 0 -1 -1 0 0 0 0 6 "
      "0 -1 0 0 0 -1 1 0 0 0 6 0 "
      "0 1 0 1 0 0 0 0 -1 3 9 9 "
      "0 -1 0 -1 0 0 0 0 -1 3 3 3 "
      "0 1 0 -1 0 0 0 0 1 9 9 3 "
      "0 -1 0 1 0 0 0 0 1 9 3 9 "
      "1 0 0 0 0 1 0 -1 0 3 9 9 "
      "-1 0 0 0 0 1 0 1 0 9 3 9 "
      "-1 0 0 0 0 -1 0 -1 0 3 3 3 "
      "1 0 0 0 0 -1 0 1 0 9 9 3 "
      "0 0 1 0 1 0 -1 0 0 3 9 9 "
      "0 0 1 0 -1 0 1 0 0 9 9 3 "
      "0 0 -1 0 1 0 1 0 0 9 3 9 "
      "0 0 -1 0 -1 0 -1 0 0 3 3 3 "
      "-1 0 0 0 -1 0 0 0 -1 6 6 6 "
      "1 0 0 0 1 0 0 0 -1 0 6 0 "
      "1 0 0 0 -1 0 0 0 1 6 0 0 "
      "-1 0 0 0 1 0 0 0 1 0 0 6 "
      "0 0 -1 -1 0 0 0 -1 0 6 6 6 "
      "0 0 -1 1 0 0 0 1 0 0 0 6 "
      "0 0 1 1 0 0 0 -1 0 0 6 0 "
      "0 0 1 -1 0 0 0 1 0 6 0 0 "
      "0 -1 0 0 0 -1 -1 0 0 6 6 6 "
      "0 1 0 0 0 -1 1 0 0 6 0 0 "
      "0 -1 0 0 0 1 1 0 0 0 0 6 "
      "0 1 0 0 0 1 -1 0 0 0 6 0 "
      "0 -1 0 -1 0 0 0 0 1 9 3 3 "
      "0 1 0 1 0 0 0 0 1 9 9 9 "
      "0 -1 0 1 0 0 0 0 -1 3 3 9 "
      "0 1 0 -1 0 0 0 0 -1 3 9 3 "
      "-1 0 0 0 0 -1 0 1 0 9 3 3 "
      "1 0 0 0 0 -1 0 -1 0 3 9 3 "
      "1 0 0 0 0 1 0 1 0 9 9 9 "
      "-1 0 0 0 0 1 0 -1 0 3 3 9 "
      "0 0 -1 0 -1 0 1 0 0 9 3 3 "
      "0 0 -1 0 1 0 -1 0 0 3 3 9 "
      "0 0 1 0 -1 0 -1 0 0 3 9 3 "
      "0 0 1 0 1 0 1 0 0 9 9 9 "
      " 0.0 0.0 0.0 0.5 0.5 0.5",
      ""};
  return reps;
}

inline spacegroup::spacegroup(int number) : num(number), polar_axes() {
  std::istringstream ss{textreps()[number]};
  ss >> name;
  auto hm_name_set = std::string{};
  getline(ss, hm_name_set, '(');
  getline(ss, hm_name_set, ')');
  auto commapos = hm_name_set.find(',');
  if (commapos != std::string::npos)
    hm_names.push_back(hm_name_set.substr(0, commapos));
  hm_names.push_back(hm_name_set.substr(commapos + 1));
  auto numop = 0;
  ss >> pointgroup >> cryst_system >> num_primitive_operators >> numop;
  operators.resize(numop);
  for (auto i = 0; i < numop; ++i) {
    ss >> operators[i][0][0];
    ss >> operators[i][0][1];
    ss >> operators[i][0][2];
    ss >> operators[i][1][0];
    ss >> operators[i][1][1];
    ss >> operators[i][1][2];
    ss >> operators[i][2][0];
    ss >> operators[i][2][1];
    ss >> operators[i][2][2];
    ss >> operators[i][0][3];
    ss >> operators[i][1][3];
    ss >> operators[i][2][3];
    operators[i][0][3] /= 12.0;
    operators[i][1][3] /= 12.0;
    operators[i][2][3] /= 12.0;
    operators[i][3][3] = 1.0;
  }
  auto ori = std::array<double, 3>{};
  while (ss >> ori[0]) {
    ss >> ori[1];
    ss >> ori[2];
    if (ori[0] < 0.0) {
      ori[0] = 0.0;
      polar_axes[0] = true;
    }
    if (ori[1] < 0.0) {
      ori[1] = 0.0;
      polar_axes[1] = true;
    }
    if (ori[2] < 0.0) {
      ori[2] = 0.0;
      polar_axes[2] = true;
    }
    origins.push_back(ori);
  }
}

inline spacegroup& spacegroup::from_number(int num) {
  static std::unique_ptr<spacegroup> groups[NumberOfSpaceGroups + 1]{};
  if (num < 1 || num > NumberOfSpaceGroups) {
    throw std::invalid_argument("invalid space group number: " +
                                std::to_string(num));
  }
  if (groups[num] == nullptr) {
    groups[num] = std::unique_ptr<spacegroup>(new spacegroup(num));
  }
  return *groups[num];
}

inline int spacegroup::symbol_to_number(std::string symbol) noexcept {
  char symcstr[10] = "";
  strncpy(symcstr, symbol.c_str(), 9);
  auto lensym = strlen(symcstr);
  for (auto p = symcstr; *p; ++p) {
    *p = (char)toupper(*p);
  }
  for (auto i = 1; i <= NumberOfSpaceGroups; ++i) {
    if (strncmp(textreps()[i], symcstr, lensym) == 0) {
      return i;
    }
  }
  return 0;
}
}
#endif
